#include "blocklist.h"
#include "cgi-search.h"
#include "cgi-indx-fils.h"
#include "cgi-index.h"
#include <ctype.h>
#include "defltlist.h"
#include <errno.h>
#include <fcntl.h>
#include "crwl-version.h"
#include <locale.h>
#include "lat-diac.h"
#include "no-alnum.h"
#include "sgent2utf.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "synonyms.h"
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include "wc2asc.h"
#include "wc2str.h"
#include "wc2num.h"
#include <wchar.h>
#include <wctype.h>

#ifdef CGS_WITH_HTTP
 #include "cgi-crawl.h"
 #include <curl/curl.h>
 #include <iconv.h>
#endif

/*
*    Crawler and indexer for websearch form.
*    Works with cgi-search.
*
*    Copyright (c) 2020 - 2023 R.J. van der Putten, Leiden, Holland, 
*    rob at sput dot nl.
* 
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License version 3 as
*    published by the Free Software Foundation.
* 
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*/

/*  Start of generic bits  */

/*
*  Global vars
*/

/*  Generate abstract file  */
int absflg;

/*  Generate compact words list file format  */
int compflg;

/*  Debug  */
int debflg;

/*  Use 32 instead of 16 bit word numbers  */
int lrgflg;

/*  Use previous wordlist as reference  */
int reflg;

/*  Use synonyms  */
int synflg;

/*  Process non-ASCII  */
int utflg;

/*
*  Wordlist properties
*
*  wordlstentr	Space available for entries
*  wordlstmax	One smaller
*  wordlstsiz	Size in bytes
*  wordlstuse	Nr of words in wordlist
*/
int wordlstentr, wordlstmax, wordlstsiz, wordlstuse, wrdidxsiz;
/*  Wordlist  */
struct wrdentr *wordlist;

/*
*  Pointers to entries in wordlist
*  Place of word in wordlist: wordlist[x]
*  For alphabetical order:    wordlist[wrdptrs[x]]
*/
int *wrdptrs;
/*  Size in bytes  */
int wordptrsiz;
/*  Compact wordlist  */
struct stridx *wrdlstidx;	/*  Index  */
char *compwordlst;		/*  Wordlist  */

/*  Word to document index  */
struct word2docs *docidx;

/*  Per doc abstract list  */

/*
*  Abstractlist properties
*
*  abstrentr	Space available for abstracts
*  abstrmax	One smaller
*  abstrsiz	Size in bytes
*/
int abstrentr, abstrmax, abstrsiz;
struct abstr *abstrlist;

/*
*  Synonym list properties
*
*  synlstadr	Memory addres of synonym list
*  synlstfd	Synonym list FD
*  synlstsiz	Synonym list size;
*  syntab	Synonym table
*  synentr	Nr of entries in Synonym table
*/
void *synlstadr;
struct synonym *syntab;
int synlstfd, synlstsiz, synentr;

#ifdef CGS_WITH_HTTP
 /*  Broken server  */
 int brokflg;

 /*  Crawl if 1  */
 int crawlflg;

 /*  Use sitmap if 1  */
 int mapflg;

 /*  Index PDF if 1  */
 int pdflg;

 /*
 *  URL list properties
 *
 *  urlstentr	Space available for entries
 *  urlstmax	Two smaller!
 *  urlstsiz	Size in bytes
 *  urlstuse	Nr of urls in urlist
 */
 int urlstentr, urlstmax, urlstsiz, urlstuse;
 struct urlentr *urlist;
 /*
 *  Pointers to entries in urlist
 *  Place of word in urllist: urlist[x]
 *  For alphabetical order:   urlist[urlptrs[x]]
 */
 int urlptrs[65536];	/*  Fixed size of 262144 bytes  */
 /*  Start list  */
 struct refurl *refurltab;
 int refurluse;

 /*
 *  External url list properties.
 *  This is just a 2D array
 *
 *  extlstentr	Space available for entries
 *  extlstmax	One smaller
 *  extlstsiz	Size in bytes
 *  extlstuse	Nr of urls in extnlst
 */
 int extlstentr, extlstmax, extlstsiz, extlstuse;
 struct exturl *extnlst;
 /*  And the pointers for alphabetical access  */
 int extptrs[65536];
 /*  Enable the report  */
 int extnflg;

 /*
 *  LibCurl stuff
 */

 /*  Curl handle  */
 CURL *curlh;
#endif


/*  Print help  */
void hlp(void)
{
    #ifdef CGS_WITH_HTTP
     fprintf(stderr, "Usage: cgi-crawl [-a] [-b] [-c] [-d] [-e] ");
     fprintf(stderr, "-f List_Of_Files [-k] [-l] [-m] [-o]\n\t");
     fprintf(stderr, "[-p] [-r] [-s] [-t] [-u] [-v] [-w Wait_Time]\n");
    #else
     fprintf(stderr, "Usage: cgi-index [-a] [-d] -f List_Of_Files [-k] [-l] ");
     fprintf(stderr, "[-o] [-r] [-s] [-t] [-u] [-v] [-w Wait_Time]\n");
    #endif
    fprintf(stderr, "'-a': Generate abstract file.\n");
    #ifdef CGS_WITH_HTTP
     fprintf(stderr, "'-b': Broken server; Use GET to emulate HEAD\n");
     fprintf(stderr, "'-c': Crawl instead of index.\n");
    #endif
    fprintf(stderr, "'-d': Enable debug.\n");
    #ifdef CGS_WITH_HTTP
     fprintf(stderr, "'-e': Enable external links report.\n");
    #endif
    fprintf(stderr, "'-k': Compact wordlist.\n");
    fprintf(stderr, "'-l': Allow more then 64k words.\n");
    #ifdef CGS_WITH_HTTP
     fprintf(stderr, "'-m': Use sitemap.\n");
    #endif
    fprintf(stderr, "'-o': Use orthography / synonyms.\n");
    #ifdef CGS_WITH_HTTP
     fprintf(stderr, "'-p': Index PDF files.\n");
    #endif
    fprintf(stderr, "'-r': Re-use old wordlist.\n");
    fprintf(stderr, "'-s': Print word stats. Warning: Long list!\n");
    fprintf(stderr, "'-t': Text output.\n");
    fprintf(stderr, "'-u': Index non-ASCII.\n");
    fprintf(stderr, "'-v': Print version and exit.\n\n");
    fprintf(stderr, "Copyright (c) 2020 - 2022 R.J. van der Putten, ");
    fprintf(stderr, "Leiden, Holland,\nrob at sput dot nl.\n");
}

/*  Print version  */
void prtvers(void)
{
    fprintf(stderr, "Version: %s\n", cgsversion);
}

/*  Free memory (and exit)  */
int freethelot(int ret)
{
    int i;

    i = 0;

    /*  Synonym list  */
    if(syntab != NULL)
	free(syntab);
    if(synlstadr != NULL)
	munmap(synlstadr, synlstsiz);
    if(synlstfd != 0)
	close(synlstfd);

    /*  Document index  */
    i = 0;
    if (docidx != NULL) {
	while (i < wordlstentr) {
	    if (docidx[i].siz > CGS_DFLT_DAT_SIZ) {
		/*  Allocated later  */
		free(docidx[i].docs);
	    }
	    i++;
	}
	docidx = NULL;
    }

    #ifdef CGS_WITH_HTTP
     if (curlh != NULL) {
	 curl_easy_reset(curlh);
	 curl_easy_cleanup(curlh);
	 curl_global_cleanup();
	 curlh = NULL;
     }
     if (extnlst != NULL) {
	 free(extnlst);
	 extnlst = NULL;
     }
     if (refurltab != NULL) {
	 free(refurltab);
	 refurltab = NULL;
     }
     if (urlist != NULL) {
	 free(urlist);
	 urlist = NULL;
     }
    #endif

    if (abstrlist != NULL) {
        free(abstrlist);
	abstrlist = NULL;
    }
    if (docidx != NULL) {
	free(docidx);
	docidx = NULL;
    }
    if (wrdptrs != NULL) {
	free(wrdptrs);
	wrdptrs = NULL;
    }
    if (wrdlstidx!= NULL) {
	free(wrdlstidx);
	wrdlstidx = NULL;
    }
    if (compwordlst != NULL) {
	free(compwordlst);
	compwordlst = NULL;
    }
    if (wordlist != NULL) {
	free(wordlist);
	wordlist = NULL;
    }

    if (ret == 0) {
	return(0);
    } else {
	/*  Error; exit  */
	exit(1);
    }
}

/*  SGML Entity compare  */
static int compent(const void *e1, const void *e2)
{
    struct sgent2u *ent1, *ent2;

    /*   Cast from void to struct ent2u   */
    ent1 = (struct sgent2u *) e1;
    ent2 = (struct sgent2u *) e2;
    /*   And then compare   */
    return(strcmp(ent1->ent, ent2->ent));
}

/*  Find SGML entity  */
int fndsgent(char *ent, char *str)
{
    struct sgent2u key, *res;

    memset(&key, 0, sizeof(key));
    res = NULL;

    if (strlen(ent) > CGS_ENT_MAX2) {
	/*  This should never happen  */
	return(0);
    }
    strcpy(key.ent, ent);

    /* void *bsearch(const void *key, const void *base, size_t nmemb, \
      size_t size,int (*compar)(const void *, const void *)); */
    res = bsearch(&key, sgent2utab, sizeof(sgent2utab) / sizeof(struct sgent2u), \
      sizeof(struct sgent2u), compent);
    if (res == NULL) {
	if(debflg != 0)
	    fprintf(stderr, "Debug: fndsgent(): entity %s not found\n", ent);
	return(0);
    } else {
	if (debflg != 0) {
	    fprintf(stderr, "Debug: fndsgent(): found entity; %s, ", ent);
	    fprintf(stderr, "converting to %s\n", res->utf);
	}
	strcat(str, res->utf);
	return(res->len);
    }
}

/*  Check for non spacing tags  */
int chktag(char *line, int pos)
{
    int i;

    i = 0;
    while (nonsptags[i].len != 0) {
	if (pos >= nonsptags[i].pos && \
	  strncasecmp(line + pos - nonsptags[i].pos, nonsptags[i].tag, \
	  nonsptags[i].len) == 0)
	    break;
	i++;
    }

    return(nonsptags[i].len);
}

/*  Remove HTML tags  */
#ifdef CGS_WITH_HTTP
 int remtag(char *in, char *out, int len, char *mhost, char *base, int *tags, \
   struct titlprps *title, int docnt, int crawl)
#else
 int remtag(char *in, char *out, int len, int *tags)
#endif
{
    int hexflg, i, j, k, ret;
    unsigned int u;
    char mbstr[8];

    #ifdef CGS_WITH_HTTP
     int m;
     char *charset;
     char link[CGS_PATH_SIZ];
     char link2[CGS_PATH_SIZ];
    #endif

    hexflg = 0;
    i = 0;
    j = 0;
    k = 0;
    ret = 0;
    u = 0;
    memset(mbstr, 0, 8);
    
    #ifdef CGS_WITH_HTTP
     m = 0;
     charset = NULL;
     memset(link,    0, CGS_PATH_SIZ);
     memset(link2,   0, CGS_PATH_SIZ);
     if (docnt > -1) {
	 /*  Crawling; We need a charset  */
	 charset = urlist[docnt].charset;
     }
    #endif

    i = 0;
    len = strlen(in);
    out[0] = 0;
    while (i < len) {
	/*  Begin tags  */
	if (strncmp(in + i, "<!--", 4) == 0) {
	    /*  Remark  */
	    *tags |= CGS_TAG_REMARK;
	    i += 3;

	/*  Tag  */
	} else if ((*tags & CGS_TAG_REMARK) == 0) {
	    /*  Not in remark  */
	    if ((*tags & CGS_TAG_RESCST) == 0) {
		/*  Not in remark, script or style  */
		if (strncasecmp(in + i, "<area", 5) == 0) {
		    /*  Imagemap area  */
		    *tags |= CGS_TAG_AREA;
		    i += 4;
		} else if (strncasecmp(in + i, "<a", 2) == 0) {
		    /*  Anchor  */
		    *tags |= CGS_TAG_ANCH;
		    i++;
		} else if ((*tags & CGS_TAG_CONT) != 0 && \
		  strncasecmp(in + i, "charset=", 8) == 0) {
		    /*  Found old style charset  */
		    i += 7;
		    #ifdef CGS_WITH_HTTP
		     if (charset != NULL && charset[0] == 0 && \
		       title != NULL && title->len == 0 && title->str[0] == 0) {
			 strncpy(charset, in + i + 1, CGS_CHARSET_MAX);
			 procharset(charset);
			 if (debflg != 0 && charset[0] != 0) {
			     fprintf(stderr, \
			       "Debug: remtag(): charset %s\n", charset);
			 }
		     }
		    #endif
		    /*  Clear content, equiv and meta  */
		    *tags &= (~ CGS_TAG_COEQME);
		} else if ((*tags & CGS_TAG_META) != 0 && \
		  strncasecmp(in + i, "charset=\"", 9) == 0) {
		    /*  Found new style charset  */
		    i += 8;
		    #ifdef CGS_WITH_HTTP
		     if (charset != NULL && charset[0] == 0 && \
		       title != NULL && title->len == 0 && title->str[0] == 0) {
			 strncpy(charset, in + i + 1, CGS_CHARSET_MAX);
			 procharset(charset);
			 if (debflg != 0 && charset[0] != 0) {
			     fprintf(stderr, \
			       "Debug: remtag(): charset %s\n", charset);
			 }
		     }
		    #endif
		    *tags &= (~ CGS_TAG_META);
		} else if ((*tags & CGS_TAG_EQUIV) != 0 && \
		  strncasecmp(in + i, "content=\"", 9) == 0) {
		    /*  Clear equiv and set content  */
		    *tags &= (~ CGS_TAG_EQUIV);
		    *tags |= CGS_TAG_CONT;
		    i += 8;
		} else if (strncasecmp(in + i, "<frame", 6) == 0 && \
		  strncasecmp(in + i, "<frameset", 9) != 0) {
		    *tags |= CGS_TAG_FRAME;
		    i += 5;
		} else if (strncasecmp(in + i, "<head>", 6) == 0) {
		    /*  Head  */
		    *tags |= CGS_TAG_HEAD;
		    i += 5;
		} else if ((*tags & CGS_TAG_OTHER) != 0 && \
		    strncasecmp(in + i, "alt=\"", 5) == 0) {
		    /*  Alt in other  */
		    *tags |= CGS_TAG_ALT;
		    i += 4;
		} else if ((*tags & CGS_TAG_ANCH) != 0) {
		    /*  In anchor  */
		    if (strncasecmp(in + i, "href=\"", 6) == 0) {
			/*  <a href; Replace anchor by href  */
			*tags &= (~ CGS_TAG_ANCH);
			*tags |= CGS_TAG_HREF;
			i += 5;
			#ifdef CGS_WITH_HTTP
			 m = 0;
			#endif
		    } else if (in[i] > ' ' || in[i] < 0) {
			/*  Not <a href; Replace anchor by other  */
			*tags &= (~ CGS_TAG_ANCH);
			*tags |= CGS_TAG_OTHER;
		    }
		} else if ((*tags & CGS_TAG_META) != 0 && \
		  strncasecmp(in + i, "http-equiv=\"Content-Type\"", 25) == 0) {
		    /*  Clear meta and set equiv  */
		    *tags &= (~ CGS_TAG_META);
		    *tags |= CGS_TAG_EQUIV;
		    i += 24;
		} else if (strncasecmp(in + i, "<iframe", 7) == 0) {
		    *tags |= CGS_TAG_IFRAME;
		    i += 6;
		} else if ((*tags & CGS_TAG_AREA) != 0) {
		    if (strncasecmp(in + i, "alt=\"", 5) == 0) {
			/*  Alt in area  */
			*tags |= CGS_TAG_ALT;
			i += 4;
		    } else if (strncasecmp(in + i, "href=\"", 6) == 0) {
			/*  Href in area  */
			*tags |= CGS_TAG_HREF;
			#ifdef CGS_WITH_HTTP
			 m = 0;
			#endif
			i += 5;
		    }
		} else if ((*tags & CGS_TAG_FRAME) != 0) {
		    if (strncasecmp(in + i, "name=\"", 6) == 0) {
			/*  Name in frame  */
			*tags |= CGS_TAG_NAME;
			i += 5;
		    } else if (strncasecmp(in + i, "src=\"", 5) == 0) {
			/*  Src in frame  */
			*tags |= CGS_TAG_SRC;
			#ifdef CGS_WITH_HTTP
			 m = 0;
			#endif
			i += 4;
		    }
		} else if ((*tags & CGS_TAG_IFRAME) != 0) {
		    if (strncasecmp(in + i, "src=\"", 5) == 0) {
			/*  Src in iframe  */
			*tags |= CGS_TAG_SRC;
			#ifdef CGS_WITH_HTTP
			 m = 0;
			#endif
			i += 4;
		    } else if (strncasecmp(in + i, "title=\"", 7) == 0) {
			/*  Title in iframe  */
			*tags |= CGS_TAG_TITLE2;
			i += 6;
		    }
		} else if ((*tags & CGS_TAG_HEAD) != 0 && \
		  strncasecmp(in + i, "<meta", 5) == 0) {
		    *tags |= CGS_TAG_META;
		    i += 4;
		} else if ((*tags & CGS_TAG_HEAD) != 0 && \
		  strncasecmp(in + i, "<title>", 7) == 0) {
		    /*  Title  */
		    *tags |= CGS_TAG_TITLE;
		    i += 6;
		    #ifdef CGS_WITH_HTTP
		     if (title != NULL ) {
			 title->str[0] = 0;
			 title->len = 0;
		     }
		    #endif
	        } else if (in[i] == '<' && \
		  strncasecmp(in + i, "<script", 7) != 0 && \
		  strncasecmp(in + i, "<style", 6) != 0) {
		    /*  Any other tag  */
		   *tags |= CGS_TAG_OTHER;
		}
	    }	/*  End of not in remark, script or style  */

	    /*  Check for script or style  */
	    if (strncasecmp(in + i, "<script", 7) == 0) {
		*tags |= CGS_TAG_SCRIPT;
		i += 6;
	    } else if (strncasecmp(in + i, "<style", 6) == 0) {
		*tags |= CGS_TAG_STYLE;
		i += 5;
	    }
	}	/*  End of not in remark  */

	/*
	*  The actual tag removal
	*  Make sure there are no '"', '<' or '>' from alt, name or title
	*/
	if (*tags == 0 || \
	  ((*tags & CGS_TAG_REMARK) == 0 && (*tags & CGS_TAG_ALNATIT2) != 0 && \
	  in[i] != '\"' && in[i] != '<' && in[i] != '>')) {
	    if (strncasecmp(in + i, "&quot;", 6) == 0) {
		i += 5;
		out[j] = '\"';
	    } else if (strncasecmp(in + i, "&amp;", 5) == 0) {
		i += 4;
		out[j] = '&';
	    } else if (strncasecmp(in + i, "&lt;", 4) == 0) {
		i += 3;
		out[j] = '<';
	    } else if (strncasecmp(in + i, "&gt;", 4) == 0) {
		i += 3;
		out[j] = '>';
	    } else if (strncasecmp(in + i, "&#xA0;", 6) == 0 || \
	      strncmp(in + i, "&#160;", 6) == 0 || \
	      strncasecmp(in + i, "&nbsp;", 6) == 0) {
		/*  nbsp, replace by space  */
		i += 5;
		out[j] = ' ';
	    } else if (strncmp(in + i, "\xC2\xA0", 2) == 0) {
		/*  UTF-8 nbsp  */
		i++;
		out[j] = ' ';
	    /*  2023-10-27 Remove Soft hyphen  */
	    } else if (strncasecmp(in + i, "&#xAD;", 6) == 0 || \
	      strncmp(in + i, "&#173;", 6) == 0) {
		i += 5;
		out[j] = 0;	/*  => No j++  */
	    } else if (strncasecmp(in + i, "&shy;", 5) == 0) {
		i += 4;
		out[j] = 0;
	    } else if (strncmp(in + i, "\xC2\xAD", 2) == 0) {
		/*  UTF-8 shy  */
		i++;
		out[j] = 0;
	    } else if (strncmp(in + i, "&#", 2) == 0) {
		/*  Numeric entity  */
		u = 0;
		mbstr[0] = 0;
		if (in[i + 2] == 'X' || in[i + 2] == 'x') {
		    /*  &#xXX; or &#xXXXX;  */
		    hexflg = 1;
		    k = 3;
		} else {
		    /*  &#DD; ... &#DDDDD;  */
		    hexflg = 0;
		    k = 2;
		}
		/*  Assumes UTF-8  */
		while (k < 8) {
		    if (in[i + k] == ';') {
			in[i + k] = 0;
			if ( ((hexflg == 0 && \
			  sscanf(in + i + 2, "%d", (int *) &u) == 1) || \
			  (hexflg != 0 && \
			  sscanf(in + i + 3, "%X", &u) == 1)) && \
			  (ret = wctomb(mbstr, (wchar_t) u)) > 0 ) {
			    mbstr[ret] = 0;	/*  Terminate  */
			    strcat(out, mbstr);
			    i += k;		/*  i++ later  */
			    j = j + ret - 1;	/*  j++ later  */
			    if (debflg != 0) {
				fprintf(stderr, \
				  "Debug: remtag(): conv nent to %s\n", \
				   mbstr);
			    }
			} else {
			    in[i + k] = ';';
			    out[j] = in[i];
			}
			break;
		    }
		    k++;
		}
	    } else if (in[i] == '&') {
		/*  Other SGML entity  */
		k = 1;
		while (k < CGS_ENT_SIZ) {
		    if (in[i + k] == ';') {
			in[i + k] = 0;
			/* int fndsgent(char *ent, char *str); */
			if ((ret = fndsgent(in + i + 1, out)) > 0) {
			    i += k;
			    j = j + ret - 1;
			    if (debflg != 0) {
				fprintf(stderr, \
				   "Debug: remtag(): conv sgml ent\n");
			    }
			} else {
			    in[i + k] = ';';
			    out[j] = in[i];
			}
			break;
		    }
		    k++;
		}
	    } else {
		out[j] = in[i];
	    }
	    /*  2023-10-27 Skip Shy  */
	    if (out[j] != 0) {
		j++;
		out[j] = 0;
	    }
	    if (j > CGS_REMTAG_OUT_MAX) {
		fprintf(stderr, "remtag(): String too long\n");
		return(-1);
	    }
	#ifdef CGS_WITH_HTTP
	    if (title != NULL && (*tags & CGS_TAG_TITLE) != 0) {
		/*  Title  */
		if (in[i] > 0 && in[i] < ' ') {
		    /*  Possable linewrap  */
		    title->str[title->len] = ' ';
		} else {
		    title->str[title->len] = in[i];
		}
		title->len++;
		title->str[title->len] = 0;
		if (title->len > CGS_PATH_MAX2) {
		    fprintf(stderr, "remtag(): Title too long\n");
		    title->len = 0;
		    return(-1);
		}
	    }
	 } else if (((*tags & (~ CGS_TAG_AREA)) == CGS_TAG_HREF || \
	   (*tags & (~ CGS_TAG_FRAME)) == CGS_TAG_SRC || \
	   (*tags & (~ CGS_TAG_IFRAME)) == CGS_TAG_SRC) && \
	   in[i] != '\"') {
	     /*  Href (possibly in area) or src in frame  */
	     link[m] = in[i];
	     m++;
	     link[m] = 0;	/*  Terminate  */
	     if (m > CGS_PATH_MAX2) {
		 fprintf(stderr, "remtag(): Url too long\n");
		 return(-1);
	     }
	#endif
	}	/*  End of tags check  */

	/*  End tags  */
	if (strncmp(in + i, "-->", 3) == 0) {
	    /*  End of remark  */
	    *tags &= (~ CGS_TAG_REMARK);
	    i += 2;
	} else if ((*tags & CGS_TAG_RESCST) == 0) {
	    /*  Not in remark, style or script  */
	    if ((*tags & CGS_TAG_ALT) != 0 && in[i] == '\"' && \
	      (in[i + 1] == '>' || (in[i + 1] < '!' && in[i + 1] > -1))) {
		/*  End of alt  */
		*tags &= (~ CGS_TAG_ALT);
		if (in[i + 1] != '>') {
		    /*  If '>' tag needs te be cleared later  */
		    i++;
		}
	    } else if ((*tags & CGS_TAG_NAME) != 0 && in[i] == '\"' && \
	      (in[i + 1] == '>' || (in[i + 1] < '!' && in[i + 1] > -1))) {
		/*  End of name  */
		*tags &= (~ CGS_TAG_NAME);
		if (in[i + 1] != '>')
		    i++;
	    } else if ((*tags & CGS_TAG_TITLE2) != 0 && in[i] == '\"' && \
	      (in[i + 1] == '>' || (in[i + 1] < '!' && in[i + 1] > -1))) {
		/*  End of title= ...  */
		*tags &= (~ CGS_TAG_TITLE2);
		if (in[i + 1] != '>')
		    i++;
	    } else if ((*tags & CGS_TAG_HRSR) != 0 && in[i] == '\"' && \
	      (in[i + 1] == '>' || (in[i + 1] == '/' && in[i + 2] == '>') || \
	      (in[i + 1] < '!' && in[i + 1] > -1))) {
		/*  End of href or src; Clear a and href or src  */
		#ifdef CGS_WITH_HTTP
		 if (debflg != 0) {
		     fprintf(stderr, \
		       "Debug: remtag(): Link: %s\n", link);
		 }
		 if (crawl != 0) {
		     /*  Note: mhost and base are NULL when crawl == 0  */
		     if (charset != NULL && charset[0] != 0) {
			 /*  URL may contain non-ASCII; Back to orig charset  */
			 link2[0] = 0;
			 /* int fromutf(char *inbuf, int inlen, char *outbuf, \
			   int outsiz, char *charset) */
			 if (fromutf(link, m, link2, CGS_PATH_MAX, charset) > 0)
			     procpath(mhost, base, link2);
			 else
			     fprintf(stderr, "remtag(): fromutf() failed\n");
		     } else {
			 /*  Current document is UTF-8  */
			 procpath(mhost, base, link);
		     }
		     /*  Urlist may have grown; Set charset pointer again  */
		     charset = urlist[docnt].charset;
		 }
		#endif
		if ((*tags & CGS_TAG_HREF) != 0) {
		    /*  Href; Clear a and href  */
		    *tags &= (~ CGS_TAG_AHREF);
		} else if ((*tags & CGS_TAG_SRC) != 0) {
		    /*  Src; Clear src  */
		    *tags &= (~ CGS_TAG_SRC);
		}
		/*
		*  New
		*
		*  <a href="Link" Some Text> will index Some Text
		*  The stuff below will prevent this
		*/
		if(*tags == 0 && in[i + 1] < '!' && in[i + 1] > -1)
		    *tags = CGS_TAG_OTHER;
		i++;
	    } else if (strncasecmp(in + i, "</a>", 4) == 0) {
		/*  Way beyond <a href; Clear to make sure  */
		*tags &= (~ CGS_TAG_AHREF);
		i += 2;		/*  Just two; We need '>' after i++  */
	    } else if (strncasecmp(in + i, "</map>", 6) == 0) {
		/*  End of imagemap; To make sure clear alt, area and href  */
		*tags &= (~ CGS_TAG_ALARHR);
		/*  Clear Other tag  */
		*tags &= (~ CGS_TAG_OTHER);
		i += 5;
	    } else if (strncasecmp(in + i, "</frameset>", 11)  == 0) {
		/*  End of frameset; To make sure clear frame, name and src  */
		*tags &= (~ CGS_TAG_FRNASR);
		/*  Clear Other tag  */
		*tags &= (~ CGS_TAG_OTHER);
		i += 10;
	    } else if (strncasecmp(in + i, "</iframe>", 9) == 0) {
		/*  End of iframe; To make sure clear iframe, src and title2  */
		*tags &= (~ CGS_TAG_IFSRT2);
		/*  Clear Other tag  */
		*tags &= (~ CGS_TAG_OTHER);
		i += 8;
	    } else if ((*tags & CGS_TAG_HEAD) != 0 && \
	      strncasecmp(in + i, "</head>", 7) == 0) {
		/*  End of head; Clear cont, equiv and meta  */
		*tags &= (~ CGS_TAG_COEQME);
		/*  End of head; Clear title  */
		*tags &= (~ CGS_TAG_TITLE);
		/*  End of head; Clear head  */
		*tags &= (~ CGS_TAG_HEAD);
		/*  Clear Other tag  */
		*tags &= (~ CGS_TAG_OTHER);
		i += 6;
	    } else if ((*tags & CGS_TAG_TITLE) != 0 && \
	      strncasecmp(in + i, "</title>", 8) == 0) {
		/*  End of title; Clear title  */
		#ifdef CGS_WITH_HTTP
		 //if (debflg != 0) {	BUG
		 if (debflg != 0 && title != NULL) {
		     fprintf(stderr, \
		       "Debug: remtag(): Title: %s\n", title->str);
		 }
		#endif
		*tags &= (~ CGS_TAG_TITLE);
		i += 7;
	    } else if (in[i] == '>') {
		/*  End of tag; Clear cont, equiv and meta  */
		*tags &= (~ CGS_TAG_COEQME);
		/*  End of tag; Clear alt, area and href  */
		*tags &= (~ CGS_TAG_ALARHR);
		/*  End of tag; Clear frame, name and src  */
		*tags &= (~ CGS_TAG_FRNASR);
		/*  End of tag; Clear iframe, src and title2  */
		*tags &= (~ CGS_TAG_IFSRT2);
		/*  Insert space, except after </a>, b, i, u  */
		if (chktag(in, i) == 0 && j > 0 && out[j - 1] != ' ') {
		    out[j] = ' ';
		    j++;
		    out[j] = 0;
		}
		/*  Clear Other tag  */
		*tags &= (~ CGS_TAG_OTHER);
	    }	/*  End of not in remark, style or script  */

	} else if ((*tags & CGS_TAG_REMARK) == 0) {
	    /*  Not in remark: Check for and of style or script  */
	    if (strncasecmp(in + i, "</script>", 9) == 0) {
		*tags &= (~ CGS_TAG_SCRIPT);
		i += 8;
	    } else if (strncasecmp(in + i, "</style>", 8) == 0) {
		*tags &= (~ CGS_TAG_STYLE);
		i += 7;
	    }
	}
	i++;
    }

    return(j);
}

/*  Increase size of docidx for all words  */
int growdocidx(int preventr)
{
    int docidxsiz, i, oldsiz;
    struct word2docs *oldptr;

    i = 0;

    oldptr = docidx;
    /*  wordlstentr is already bigger  */
    docidxsiz = wordlstentr * sizeof(struct word2docs);
    oldsiz = docidxsiz / 2;

    docidx = (struct word2docs *) realloc(docidx, docidxsiz);
    if (docidx == NULL) {
	fprintf(stderr, "growdocidx(): Can't allocate memory for docidx.\n");
	freethelot(1);
    }
    memset(docidx + preventr, 0, oldsiz);

    if (docidx != oldptr) {
	/*  New address  */
	for (i = 0; i < preventr; i++) {
	    if (docidx[i].siz <= CGS_DFLT_DAT_SIZ) {
		/*  Data is in struct  */
		docidx[i].docs = &(docidx[i].data[0]);
	    }
	}
    }

    /*  fill new with defaults  */
    i = preventr;
    while (i < wordlstentr) {
	docidx[i].siz  = CGS_DFLT_DAT_SIZ;
	docidx[i].docs = &(docidx[i].data[0]);
	i++;
    }

    if (debflg != 0) {
	fprintf(stderr, \
	  "Debug: growdocidx(): Growing docidx from %d to %d bytes at %lx\n", \
	  oldsiz, docidxsiz, (unsigned long) docidx);
    }

    return(0);
}

/*  Increase size of compact wordlist  */
void growcompwrdlst(void)
{
    int size;

    size = 0;

    /*  Double the size of wordlst. Check maximum first  */
    if (wordlstsiz > 1073741820) {
	fprintf(stderr, "growcompwrdlst(): Wordlist too large\n");
	freethelot(1);
    }

    size = 2 * wordlstsiz;
    if(debflg != 0)
	fprintf(stderr, "Debug: growcompwrdlst(): Old: %d New: %d bytes\n", \
	  wordlstsiz, size);

    compwordlst = (char *) realloc(compwordlst, (size_t) size);
    if (compwordlst == NULL) {
	fprintf(stderr, "growcompwrdlst(): Can't allocate memory for wordlist\n");
	freethelot(1);
    }
    memset(compwordlst + wordlstsiz, 0, wordlstsiz);

    if (debflg != 0) {
	fprintf(stderr, \
	  "Debug: growcompwrdlst(): Growing wordlist from %d to %d bytes at %lx\n", \
	  wordlstsiz, size, (unsigned long) compwordlst);
    }
    wordlstsiz = size;
}

/*  Increase size of wrdlstidx  */
int growrdlstidx(int preventr)
{
    int size;

    size = 0;

    /*  Double the size of wrdlstidx. Check maximum first  */
    if (wrdidxsiz > 1073741820) {
	fprintf(stderr, "growrdlstidx(): Word list index too large\n");
	freethelot(1);
    }
    size = 2 * wrdidxsiz;
    wrdlstidx = (struct stridx *) realloc(wrdlstidx, size);
    if (wrdlstidx == NULL) {
	fprintf(stderr, "growrdlstidx(): Can't allocate memory for wrdlstidx\n");
	freethelot(1);
    }
    /*  Pointer arithmatic  */
    memset(wrdlstidx + preventr, 0, wrdidxsiz);
    if (debflg != 0) {
	fprintf(stderr, \
	  "Debug: growrdlstidx(): Growing wrdlstidx from %d to %d bytes at %lx\n", \
	  wrdidxsiz, size, (unsigned long) wrdlstidx);
    }
    wrdidxsiz = size;

    return(0);
}

/*  Increase size of wrptrs  */
int growrdptrs(int preventr)
{
    int size;

    size = 0;

    /*  Double the size of wordptrs. Check maximum first  */
    if (wordptrsiz > 1073741820) {
	fprintf(stderr, "growrdptrs(): Word pointers too large\n");
	freethelot(1);
    }
    size = 2 * wordptrsiz;
    wrdptrs = (int *) realloc(wrdptrs, size);
    if (wrdptrs == NULL) {
	fprintf(stderr, "growrdptrs(): Can't allocate memory for wrptrs\n");
	freethelot(1);
    }
    memset(wrdptrs + preventr, 0, wordptrsiz);
    if (debflg != 0) {
	fprintf(stderr, \
	  "Debug: growrdptrs(): Growing wrdptrs from %d to %d bytes at %lx\n", \
	  wordptrsiz, size, (unsigned long) wrdptrs);
    }
    wordptrsiz = size;

    return(0);
}

/*  Increase size of wordlist  */
int growrdlst(int preventr)
{
    int size;

    size = 0;

    /*  Double the size of wordlst. Check maximum first  */
    if (wordlstsiz > 1073741820) {
	fprintf(stderr, "growrdlst(): Wordlist too large\n");
	freethelot(1);
    }
    size = 2 * wordlstsiz;
    wordlist = (struct wrdentr *) realloc(wordlist, size);
    if (wordlist == NULL) {
	fprintf(stderr, "growrdlst(): Can't allocate memory for wordlist\n");
	freethelot(1);
    }
    /*  Pointer arithmatic  */
    memset(wordlist + preventr, 0, wordlstsiz);
    if (debflg != 0) {
	fprintf(stderr, \
	  "Debug: growrdlst(): Growing wordlist from %d to %d bytes at %lx\n", \
	  wordlstsiz, size, (unsigned long) wordlist);
    }
    wordlstsiz  = size;

    return(0);
}

/*  Grow various tables  */
int growtables(int preventr)
{
    /*  Grow wordlist  */
    if (compflg == 0) {
	growrdlst(preventr);
    } else {
	/*  Compact worlist  */
	growrdlstidx(preventr);
    }

    /*  Grow of wrptrs  */
    growrdptrs(preventr);

    /*  Double the size of wordlstentr  */
    wordlstentr = wordlstentr * 2;
    wordlstmax  = wordlstentr - 1;

    return(0);
}

/*  Convert word number to word  */
char * idx2word(int i)
{
    char *word;

    word = NULL;

    /*  Note: i = num -1 !!!  */
    if (wrdlstidx[i].ptr + wrdlstidx[i].len > wordlstsiz) {
	fprintf(stderr, "idx2word(): Wordlist index mismatch\n");
	freethelot(1);
    }
    word = compwordlst + wrdlstidx[i].ptr;

    return(word);
}

/*  Add word to word list  */
int addword(char *str, int len, int compwrdstat)
{
    int i, j, num, preventr;
    char *word;

    i = 0;
    j = 0;
    num = 0;
    preventr = 0;
    word = NULL;

    /*
    *  wordlist[x]		Order is added words
    *  wordlist[wrdptrs[x]]	Order is alphabetical
    */

    /*  compwrdstat is status of wrdsrch()  */
    i = compwrdstat;
    if(i < 0)
	i = 0;

    if (compflg == 0) {
	word = wordlist[wrdptrs[i]].str;
    } else {
	/*  Compact wordlist  */
	word = idx2word(wrdptrs[i]);
    }

    if (reflg != 0 && debflg != 0) {
	fprintf(stderr, "Debug: addword(): Start at %s %04X alph seq %04X\n", \
	  word, wrdptrs[i], i);
    }

    while (i < wordlstuse) {
	if(compflg == 0)
	    word = wordlist[wrdptrs[i]].str;
        else
	    word = idx2word(wrdptrs[i]);
	if (strcmp(str, word) < 0) {
	    /*  Str < wordlist[i].str  */
	    j = i;	/*  To be pos of str  */
	    break;
	}
	i++;
    }

    if (i == wordlstuse) {
	/*  New entry all the way at the back  */
	wrdptrs[wordlstuse] = wordlstuse;
	j = wordlstuse; /*  For debug  */
    } else {
	i = wordlstuse;
	while (i > j) {
	    wrdptrs[i] = wrdptrs[i - 1];
	    i--;
	}
	/*  Other new entry  */
	wrdptrs[j] = wordlstuse;
    }
    /*  Common stuff  */
    if (compflg == 0) {
	strcpy(wordlist[wordlstuse].str, str);
	wordlist[wordlstuse].num = wordlstuse + 1;
    } else {
	if (wordlstuse == 0) {
	    /*  1st entry  */
	    wrdlstidx[0].ptr = 0;
	    word = compwordlst;
	} else {
	    wrdlstidx[wordlstuse].ptr = \
	      wrdlstidx[wordlstuse - 1].ptr + wrdlstidx[wordlstuse - 1].len + 1;
	    word = compwordlst + wrdlstidx[wordlstuse].ptr;
	}
	strcpy(word, str);
	word[len] = 0;  /*  Terminate  */
	wrdlstidx[wordlstuse].num = wordlstuse + 1;
	wrdlstidx[wordlstuse].len = len;
    }

    /*  Check  */
    if (reflg != 0 && debflg != 0) {
	if (compflg == 0) {
	    word = wordlist[wordlstuse].str;
	    num  = wordlist[wordlstuse].num;
	} else {
	    word = compwordlst + wrdlstidx[wordlstuse].ptr;
	    num  = wrdlstidx[wordlstuse].num;
	}
	fprintf(stderr, "Debug: addword(): added word %s at %04X alph seq %04X\n", \
	  word, num, j + 1);
    }
    wordlstuse++;

    /*  Check wordlstuse size and realloc if needed  */
    if (wordlstuse > wordlstmax) {
	if (lrgflg == 0 && wordlstentr > 65530) {
	    /*  16 bit word numbers  */
	    fprintf(stderr, "addword(): Number of words too large\n");
	    freethelot(1);
	}

	/*  Store previous nr of entries  */
	preventr = wordlstentr;
	/*  Grow various tables  */
	growtables(preventr);
	/*  Grow the word to doc index as well  */
	growdocidx(preventr);
    }

    /*  Separate check for compact worldlist  */
    if (compflg != 0) {
	/*
	*  Foobar0CGS_WRD_SI0
	*  ^     ^^ ptr
	*  ptr-- len--
	*/
	if (wrdlstidx[wordlstuse - 1].ptr + wrdlstidx[wordlstuse - 1].len + \
	  CGS_WRD_SIZ >= wordlstsiz) {
	    growcompwrdlst();
	}
    }

    return(0);
}

/*  Increase size of docidx for one word  */
int growidxdat(int num)
{
    int oldsiz, size;
    unsigned short *docs;

    oldsiz = 0;
    size = 0;
    docs = NULL;

    if (docidx[num].siz <= CGS_DFLT_DAT_SIZ) {
	/*  First time  */
	size = CGS_NEXT_DAT_SIZ * sizeof(unsigned short);
	docs = malloc((size_t) size);
	if (docs == NULL) {
	    fprintf(stderr, \
	      "growidxdat() Can't allocate memory for docidx %04X\n", num);
	    freethelot(1);
	}
	memset(docs, 0, size);
	if (debflg != 0) {
	    fprintf(stderr, \
	      "Debug: growidxdat(): Growing docidx %04X from %d to %d\n", \
	      //num, docidx[num].siz, (int) (size / sizeof(unsigned short)));
	      num, CGS_DFLT_DAT_SIZ, CGS_NEXT_DAT_SIZ);
	}
	memcpy(docs, docidx[num].docs, \
	  docidx[num].siz * sizeof(unsigned short));
	docidx[num].docs = docs;
	//docidx[num].siz = size / sizeof(unsigned short);
	docidx[num].siz = CGS_NEXT_DAT_SIZ;
    } else {
	/*  Double in size  */
	oldsiz = docidx[num].siz * sizeof(unsigned short);
	size = oldsiz * 2;
	docidx[num].docs = (unsigned short *) realloc(docidx[num].docs, size);
	if (docidx[num].docs == NULL) {
	    fprintf(stderr, \
	      "growidxdat() Can't allocate memory for docidx %04X\n", num);
	    freethelot(1);
	}

	memset(docidx[num].docs + docidx[num].siz, 0, oldsiz);
	if (debflg != 0) {
	    fprintf(stderr, \
	      "Debug: growidxdat(): Growing docidx %04X from %d to %d\n", \
	      num, docidx[num].siz, docidx[num].siz * 2);
	}
	docidx[num].siz = docidx[num].siz * 2;
    }

    return(0);
}

/*  Add word nr to abstract  */
int addtoabs(int wrdnr, int docnr)
{
    if (abstrlist[docnr].words[CGS_ABSTR_MAX2] != 0) {
	/*  Extra check  */
	return(-1);
    }

    if (abstrlist[docnr].words[0] == 0) {
	/*  For debug and checks  */
	abstrlist[docnr].words[0] = docnr + 1;
	/*  Ptr  */
	abstrlist[docnr].words[CGS_ABSTR_MAX] = 1;
    }

    /*  Use abstrlist[docnr].words[CGS_ABSTR_MAX] as ptr  */
    abstrlist[docnr].words[abstrlist[docnr].words[CGS_ABSTR_MAX]] = wrdnr + 1;
    abstrlist[docnr].words[CGS_ABSTR_MAX]++;
    if (abstrlist[docnr].words[CGS_ABSTR_MAX] > CGS_ABSTR_MAX2) {
	/*  Terminate  */
	abstrlist[docnr].words[CGS_ABSTR_MAX] = 0;
    }

    return(0);
}

/*  Bsearch like search in wordlist  */
int wrdsrch(char *key, int *compwrdstat)
{
    int first, last, middle, ret;
    char *word;

    first = 0;
    ret = 0;
    word = NULL;

    last = wordlstuse - 1;
    middle = (first + last) / 2;

    while (first <= last) {
	if (compflg == 0) {
	    word = wordlist[wrdptrs[middle]].str;
	} else {
	    /*  Compact wordlist  */
	    word = idx2word(wrdptrs[middle]);
	}
	ret = strcmp(word, key);
	if (ret < 0) {
	    /*  str < key. Examine 2nd half  */
	    first = middle + 1;
	} else if (ret == 0) {
	    /*  str == key; Found  */
	    return(wrdptrs[middle]);
	} else {
	    /*  str > key. Examine 1st half  */
	    last = middle - 1;
	}
	middle = (first + last) / 2;
    }

    *compwrdstat = middle;
    if (ret > 0) {
	/*  Too far  */
	(*compwrdstat)--;
    }

    return(-1);
}

/*  Lookup word in wordlist. This calls addword() if word is new.  */
int findword(char *str, int len, int docnt, int mode)
{
    int  compwrdstat, i, next, num;
    char *key = str;
    int  res;

    compwrdstat = 0;
    i = 0;
    next = 0;
    num = 0;
    res = 0;

    /* int wrdsrch(char *key, int *compwrdstat) */
    res = wrdsrch(key, &compwrdstat);
    if (res < 0 ) {
	/*  Word not found, add to list  */
	addword(str, len, compwrdstat);
	num = wordlstuse - 1;
    } else {
	/*  Word found  */
	if (compflg == 0)
	    num = wordlist[res].num - 1;
	else
	    num = wrdlstidx[res].num - 1;
    }

    /*  Nr and word are known  */
    if (absflg != 0 && mode != 0 && abstrlist[docnt].words[CGS_ABSTR_MAX2] == 0)
	addtoabs(num, docnt);

    /*  Add document nr to list  */
    next = docidx[num].next;		/*  Next free pos in list  */
    i = 0;
    while (i <= next) {
	if(docidx[num].docs[i] == docnt + 1)	/*  Starts numbering at 1  */
	    /*  Already in list  */
	    return(0);
	i++;
    }

    docidx[num].docs[next] = (unsigned short) (docnt + 1);
    docidx[num].next++;

    if(docidx[num].next >= docidx[num].siz)
	growidxdat(num);

    return(0);
}

/*  Compare blockword  */
static int fndblockwrd(const void *w1, const void *w2)
{
    char *wrd1, *wrd2;

    wrd1 = (char *) w1;
    wrd2 = (char *) w2;

    return(strcmp(wrd1, wrd2));
}

/*  Check word blocklist  */
int chkblockwrd(char *str)
{
    char *key = str;
    char *res;

    res = NULL;

    /* void *bsearch(const void *key, const void *base, size_t nmemb, \
      size_t size,int (*compar)(const void *, const void *)); */
    res = bsearch(key, blocklst, sizeof(blocklst) / sizeof(blocklst[0]), \
      sizeof(blocklst[0]), fndblockwrd);

    if (res == NULL) {
	return(0);
    } else {
	return(-1);
    }
}

/*  Find ASCII words in string  */
int findawords(char *str, int len, int docnt, int *chrcnt)
{
    int i, j, strsiz, wrdflg;

    i = 0;
    j = 0;
    strsiz = 0;
    wrdflg = 0;

    //while (i < len) {	Changed to get NULL char to flip alnum state
    while (i <= len) {
	/*  Finds tekst  */
	if(str[i] > '@' && str[i] < ']')
	    str[i] = tolower(str[i]);
	if ((str[i] > '/' && str[i] < ':') || \
	  (str[i] > '`' && str[i] < '{') || \
	  (i > 0 && (str[i] == '.' || str[i] == ':') && \
	  isalnum((unsigned char) str[i - 1]) != 0 && \
	  isalnum((unsigned char) str[i + 1]) != 0) || \
	  (str[i] == ':' && str[i + 1] == ':') || \
	  (i > 0 && str[i - 1] == ':' && str[i] == ':') || \
	  (utflg != 0 && str[i] < 0)) {
	    if (wrdflg == 0) {
		j = i;
		wrdflg = 1;
	    }
	} else {
	    if (wrdflg == 1) {
		str[i] = 0;
		strsiz = i - j;
		if (strsiz > CGS_MIN_WRD_LEN && strsiz < CGS_WRD_SIZ && \
		  chkblockwrd(str + j) == 0)
		    findword(str + j, strsiz, docnt, 1);
		wrdflg = 0;
		j = i + 1;
		if(strsiz > *chrcnt)
		    *chrcnt = strsiz;
	    }
	}
	i++;
    }

    return(0);
}

/*  wchar_t compare  */
static int compwchar(const void *wc1, const void *wc2)
{
    wchar_t *wchar1, *wchar2;

    /*   Cast from void to wchar_t   */
    wchar1 = (wchar_t *) wc1;
    wchar2 = (wchar_t *) wc2;
    /*   And then compare   */
    return(*wchar1 - *wchar2);
}

/*  Lookup non printable non ASCII  */
int chkhigh(wchar_t wchar)
{
    wchar_t key, *res;

    res = NULL;
    key = wchar;

    /* void *bsearch(const void *key, const void *base, size_t nmemb, \
      size_t size,int (*compar)(const void *, const void *)); */
    res = bsearch(&key, (void *) nonalnum, sizeof(nonalnum) / sizeof(wchar_t), \
       sizeof(wchar_t), compwchar);
    if (res == NULL)
	return(0);
    else
	return(*res);
}

/*  Lookup ASCII replacement for non ASCII  */
static int compwasc(const void *wa1, const void *wa2)
{
    struct wc2asc *wasc1, *wasc2;

    wasc1 = (struct wc2asc *) wa1;
    wasc2 = (struct wc2asc *) wa2;

    return(wasc1->wc - wasc2->wc);
}

/*  Convert wide-char non-ASCII to char ASCII  */
int wctoasc(wchar_t *wstr, char *str, int mode)
{
    int i, j, size;
    struct wc2asc key, *res, *wctab;

    i = 0;
    j = 0;
    size = 0;
    memset(&key, 0, sizeof(key));
    res = NULL;
    wctab = NULL;

    if (mode == 0) {
	/*  To ASCII  */
	wctab = wc2strtab;
	size = sizeof(wc2strtab) / sizeof(struct wc2asc);
    } else {
	/*  To number  */
	wctab = wc2numtab;
	size = sizeof(wc2numtab) / sizeof(struct wc2asc);
    }

    while (wstr[i] != 0) {
	if (wstr[i] > 127) {
	    /*  non-ASCII  */
	    key.wc = wstr[i];
	    /* void *bsearch(const void *key, const void *base, size_t nmemb, \
	      size_t size,int (*compar)(const void *, const void *)); */
	    res = bsearch(&key, wctab, size, sizeof(struct wc2asc), \
	      compwasc);
	    if (res == NULL) {
		/*  Not found  */
		return(-1);
	    }
	    if (debflg != 0) {
		fprintf(stderr, "Debug: wctoasc(): Found %s for %lc\n", \
		  res->str, wstr[i]);
	    }
	    if (j + res->len > CGS_WRD_MAX2) {
		/*  Too long  */
		return(-1);
	    }
	    strcat(str, res->str);
	    j += res->len;
	} else {
	    str[j] = (char) (0xFF & wstr[i]);
	    j++;
	    str[j] = 0;
	    if (j > CGS_WRD_MAX2) {
		/*  Too long  */
		return(-1);
	    }
	}
	i++;
    }

    return(j);
}

/*  Diacritcals  */
int findiac(wchar_t *wstr, char *str)
{
    int i, j;
    wchar_t key, *res;

    i = 0;
    j = 0;
    res = NULL;

    while (wstr[i] != 0) {
	if (wstr[i] < 128) {
	    str[j] = (char) (0xFF & wstr[i]);
	    j++;
	    str[j] = 0;
	    if(j > CGS_WRD_MAX2)
		/*  Too long  */
		return(-1);
	} else {
	    key = wstr[i];
	    res = bsearch(&key, (void *) diactab, \
	      sizeof(diactab) / sizeof(wchar_t), sizeof(wchar_t), compwchar);
	    if (res == NULL)
		/*  Not a latin diacritic  */
		return(-1);
	}
	i++;
    }

    return(j);
}

/*  Compare synonyms  */
static int compsyn(const void *wrd1, const void *wrd2)
{
    struct synonym *syn1, *syn2;

    /*   Cast from void to struct synom   */
    syn1 = (struct synonym *) wrd1;
    syn2 = (struct synonym *) wrd2;

    /*   And then compare   */
    return(strcmp(syn1->org, syn2->org));
}

/*  Find synonyms  */
char * findsym(char *str)
{
    struct synonym key, *res;

    res = NULL;
    memset(&key, 0, sizeof(key));

    key.org = str;
    /*
    res = bsearch(&key, synomlst, sizeof(synomlst) / sizeof(synomlst[0]), \
      sizeof(synomlst[0]), compsym);
    */
    res = bsearch(&key, syntab, synentr, sizeof(struct synonym), compsyn);
    if (res != NULL) {
	/*  res->syn is entry in synomlst  */
	if (debflg != 0) {
	    fprintf(stderr, \
	      "Debug: findsym(): Found synonym %s for %s\n", res->syn, str);
	}
	return(res->syn);
    }

    return(NULL);
}

/*  Find UTF-8 words in string  */
int finduwords(char *str, int len, int docnt, int *chrcnt)
{
    int i, j, mbstrsiz, ret, wlen, wrdflg, wstrsiz;

    char    ascstr[CGS_WRD_SIZ], mbstr[CGS_MBS_SIZ], *synom;;
    wchar_t wstr[CGS_BUF_SIZ];

    i = 0;
    j = 0;
    mbstrsiz = 0;	/*  UTF-8 string length  */
    ret = 0;
    wrdflg = 0;
    wlen = 0;		/*  Wide char string length  */
    wstrsiz = 0;
    synom = NULL;
    memset(ascstr, 0, sizeof(ascstr));
    memset(mbstr,  0, sizeof(mbstr));
    memset(wstr,   0, sizeof(wstr));

    /* size_t mbstowcs(wchar_t *dest, const char *src, size_t n); */
    wlen = mbstowcs(wstr, str, CGS_BUF_MAX);
    if (wlen < 0) {
	fprintf(stderr, "finduwords(): mbstowcs failed in doc %04X\n", docnt + 1);
	return(-1);
    }

    /*  Force all tolower  */
    for (i = 0; i < wlen; i++) {
	if (wstr[i] != 0x2126) {
	    /*  OHM SIGN is not an GREEK CAPITAL LETTER OMEGA  */
	    wstr[i] = towlower(wstr[i]);
	}
    }

    i = 0;
    //while (i < wlen) {	Changed to get NULL char to flip alnum state
    while (i <= wlen) {
 	/*
	*  Finds tekst. iswalnum() sux a bit. So it's all DIY;
	*  0-9, a-z, alnum.alnum, alnum:alnum, ::
	*  > 160 except graphics, quotes and spaces.
	*/
	/*  [0-9] || [a-z]  */
	if ((wstr[i] > '/' && wstr[i] < ':') || \
	  (wstr[i] > '`' && wstr[i] < '{') || \

	  /*  alnum.alnum || alnum:alnum  */
          (i > 0 && (wstr[i] == '.' || wstr[i] == ':') && \
	  ((wstr[i - 1] > '/' && wstr[i - 1] < ':') || \
	   (wstr[i - 1] > '`' && wstr[i - 1] < '{')) && \
	  ((wstr[i + 1] > '/' && wstr[i + 1] < ':') || \
	   (wstr[i + 1] > '`' && wstr[i + 1] < '{'))) || \

	  /*  ::  */
          (wstr[i] == ':' && wstr[i + 1] == ':') || \
          (i > 0 && wstr[i - 1] == ':' && wstr[i] == ':') || \

	  (wstr[i] > 160 && chkhigh(wstr[i]) == 0)) {
            if (wrdflg == 0) {
                j = i;
                wrdflg = 1;
            }
        } else {
            if (wrdflg == 1) {
                wstr[i] = 0;
                wstrsiz = i - j;
                if (wstrsiz > CGS_MIN_WRD_LEN && wstrsiz < CGS_WRD_SIZ) {
		    /* size_t wcstombs(char *dest, const wchar_t *src, size_t n); */
		    mbstrsiz = wcstombs(mbstr, wstr + j, CGS_MBS_MAX);
		    if (mbstrsiz < 0) {
			fprintf(stderr, "finduwords(): wcstombs failed in doc %04X\n", \
			  docnt + 1);
			return(-1);
		    } else if (mbstrsiz > CGS_MIN_WRD_LEN && \
		      mbstrsiz < CGS_WRD_SIZ && chkblockwrd(mbstr) == 0) {
			findword(mbstr, mbstrsiz, docnt, 1);
			/*  Check ASCII translation  */
			ascstr[0] = 0;
			if ((ret = wctoasc(wstr + j, ascstr, 0)) > 0 && \
			  strcmp(mbstr, ascstr) != 0 && \
			  chkblockwrd(ascstr) == 0) {
			    if (debflg != 0) {
				fprintf(stderr, \
				  "Debug: finduwords(): adding %s\n", ascstr);
			    }
			    /*  Don't add to abstract  */
			    findword(ascstr, ret, docnt, 0);
			}
			/*  And numerals  */
			ascstr[0] = 0;
			if ((ret = wctoasc(wstr + j, ascstr, 1)) > 0 && \
			  strcmp(mbstr, ascstr) != 0) {
			    if (debflg != 0) {
				fprintf(stderr, \
				  "Debug: finduwords(): adding %s\n", ascstr);
			    }
			    /*  Don't add to abstract  */
			    findword(ascstr, ret, docnt, 0);
			}
			/*  Check for combining  */
			if ((ret = findiac(wstr + j, ascstr)) > 0 && \
			  strcmp(mbstr, ascstr) != 0) {
			    if (debflg != 0) {
				fprintf(stderr, \
				  "Debug: finduwords(): adding %s\n", ascstr);
			    }
			    /*  Don't add to abstract  */
			    findword(ascstr, ret, docnt, 0);
			}
			/*  2023-10-16; Check for synonyms  */
			if (synflg != 0 && syntab != NULL && \
			  (synom = findsym(mbstr)) != NULL && \
			  strcmp(mbstr, synom) != 0) {
			    if (debflg != 0) {
				fprintf(stderr, \
				  "Debug: finduwords(): adding %s\n", synom);
			    }
			    /*  Don't add to abstract  */
			    findword(synom, ret, docnt, 0);
			}
		    }
		}
                wrdflg = 0;
                j = i + 1;

                if(mbstrsiz > *chrcnt)
                    *chrcnt = mbstrsiz;
            }
        }
	i++;
    }

    return(0);
}

/*  Find words in string  */
int findwords(char *str, int len, int docnt, int *chrcnt)
{
    if(utflg == 0)
	return(findawords(str, len, docnt, chrcnt));
    else
	return(finduwords(str, len, docnt, chrcnt));
}

/*  Dump word list  */
int dumpwordlist(int mode)
{
    char *str;
    int i, k, num, wrdcnt;
    struct stridx index;
    FILE *fidx, *fout;

    /*
    mode 0 is binary, 1 is text
    wordlist[i] is sequential
    wordlist[wrdptrs[i]] is alphabetical
    */

    i = 0;
    k = 0;
    num = 0;
    memset(&index, 0, sizeof(index));
    fidx = NULL;
    fout = NULL;
    str = NULL;

    wrdcnt = 1;	/*  Start numbering at 1  */

    if (mode == 0) {
	/*  words-list  */
	if ((fout = fopen(CGS_WORDS_LIST, "w")) == NULL) {
	    fprintf(stderr, "dumpwordlist(): Can't open %s\n", CGS_WORDS_LIST);
	    return(-1);
        }
	if (compflg == 0) {
	    /*  To avoid confusion, remove words.idx if present  */
	    /* int faccessat(int dirfd, const char *pathname, int mode, int flags); */
	    if (faccessat(AT_FDCWD, CGS_WORDS_IDX, F_OK | R_OK | W_OK, \
	      0) == 0) {
		if (unlink(CGS_WORDS_IDX) < 0) {
		    fprintf(stderr, "dumpwordlist(): Can't remove %s\n", \
		      CGS_WORDS_IDX);
		    fclose(fout);
		    return(-1);
		}
		fprintf(stderr, "dumpwordlist(): Removed %s\n", CGS_WORDS_IDX);
	    }
	} else {
	    /*  Compact wordlist; Need words.idx  */
	    if ((fidx = fopen(CGS_WORDS_IDX, "w")) == NULL) {
		fprintf(stderr, "dumpwordlist(): Can't open %s\n", \
		  CGS_WORDS_IDX);
		fclose(fout);
		return(-1);
	    }
	}
    } else {
	/*  num-words.list  */
	if ((fout = fopen(CGS_WORDS_TXT, "w")) == NULL) {
	    fprintf(stderr, "dumpwordlist(): Can't open %s\n", CGS_WORDS_TXT);
	    return(-1);
	}
    }

    while (i < wordlstuse) {
	if (compflg == 0) {
	    k = wordlist[wrdptrs[i]].num;
	    /*  Renumber  */
	    wordlist[wrdptrs[i]].num = wrdcnt;
	} else {
	    /*  Compact  */
	    k = wrdlstidx[wrdptrs[i]].num;
	    wrdlstidx[wrdptrs[i]].num = wrdcnt;
	}

	if (docidx[k - 1].next != 0) {
	    /*  Is used  */
	    if (compflg == 0) {
		str = wordlist[wrdptrs[i]].str;
		num = wordlist[wrdptrs[i]].num;
		if (mode == 0) {
		    /*  Binary  */
		    fwrite(&(wordlist[wrdptrs[i]]), sizeof(struct wrdentr), 1, fout);
		} else {
		    /*  Text  */
		    if (lrgflg == 0) {
			fprintf(fout, "%04X %s\n", num, str);
		    } else {
			fprintf(fout, "%08X %s\n", num, str);
		    }
		}
	    } else {
		/*  Compact  */
		str = compwordlst + wrdlstidx[wrdptrs[i]].ptr;
		num = wrdlstidx[wrdptrs[i]].num;
		if (mode == 0) {
		    /*  Binary  */
		    index.num = num;
		    index.len = wrdlstidx[wrdptrs[i]].len;
		    /*  Word  */
		    fwrite(str, 1, index.len + 1, fout);
		    /*  Index  */
		    fwrite(&index, sizeof(struct stridx), 1, fidx);
		    /*  Next pointer  */
		    index.ptr = index.ptr + index.len + 1;
		} else {
		    /*  Text  */
		    if (lrgflg == 0) {
			fprintf(fout, "%04X %s\n", num, str);
		    } else {
			fprintf(fout, "%08X %s\n", num, str);
		    }
		}
	    }
	    wrdcnt++;
	} else if (debflg != 0) {
	    if (compflg == 0) {
		fprintf(stderr, \
		  "Debug: dumpwordlist(): removed word %s nr %04X\n", \
		  wordlist[wrdptrs[i]].str, k);
	    } else {
		fprintf(stderr, \
		  "Debug: dumpwordlist(): removed word %s nr %04X\n", \
		  compwordlst + wrdlstidx[wrdptrs[i]].ptr, k);
	    }
	}
	i++;
    }
    printf("dumpwordlist(): Wrote %d words to wordlist\n", wrdcnt - 1);

    if(fidx != NULL)
	fclose(fidx);
    fclose(fout);

    return(0);
}

/*  Dump document to word index  */
int dumpindex(int mode)
{
    int i, idxcnt, j;
    unsigned short u;
    FILE *f1, *f2;
    struct stridx strindex;

    /*  mode 1 is text  */

    i = 0;
    j = 0;
    u = 0;
    f1 = NULL;
    f2 = NULL;
    memset(&strindex, 0, sizeof(strindex));

    idxcnt = 1;

    if (mode == 0) {
	/*  index.idx  */
	if ((f1 = fopen(CGS_INDEX_IDX, "w")) == NULL) {
	   fprintf(stderr, "dumpindex(): Can't open %s\n", CGS_INDEX_IDX);
	   return(-1);
	}
	/*  index-list  */
	if ((f2 = fopen(CGS_INDEX_LIST, "w")) == NULL) {
	   fprintf(stderr, "dumpindex(): Can't open %s\n", CGS_INDEX_LIST);
	   fclose(f1);
	   return(-1);
	}
    } else {
	/*  index.list  */
	if ((f1 = fopen(CGS_INDEX_TXT, "w")) == NULL) {
	    fprintf(stderr, "dumpindex(): Can't open %s\n", CGS_INDEX_TXT);
	    return(-1);
	}
    }

    i = 0;
    while (i < wordlstuse) {
	if (docidx[wrdptrs[i]].next == 0) {
	    /*  Don't write the index  */
	    i++;
	    continue;
	}
	if (mode == 0) {
	    strindex.num = idxcnt;
	    /*  Lower 16 bit of num   */
	    u = (unsigned short) (0xFFFF & strindex.num);
	    fwrite(&u, 2, 1, f2);
	    /*  Documents  */
	    j = 0;
	    while (j < docidx[wrdptrs[i]].next) {
		fwrite(&(docidx[wrdptrs[i]].docs[j]), 2, 1, f2);
		j++;
	    }
	    u = 0;
	    fwrite(&u, 2, 1, f2);		/*  Terminate  */
	    /*  Index  */
	    strindex.len = 2 * j;		/*  In bytes  */
	    fwrite(&strindex, sizeof(strindex), 1, f1);
	    strindex.ptr = strindex.ptr + strindex.len + 4;
	    /*  Check maximum  */
	    if (strindex.ptr > 2000000000) {
		fprintf(stderr, "dumpindex(): index.idx too large\n");
		freethelot(1);
	    }
	} else {
	    fprintf(f1, "%04X ", (0xFFFF & idxcnt));
	    j = 0;
	    /*  Renumber  */
	    while (j < docidx[wrdptrs[i]].next) {
		fprintf(f1, "%04X", docidx[wrdptrs[i]].docs[j]);
		j++;
		if(docidx[wrdptrs[i]].docs[j] != 0)
		    fprintf(f1, " ");
	    }
	    fprintf(f1, "\n");
	}
	i++;
	idxcnt++;
    }

    printf("dumpindex(): Wrote %d indexes to indexlist\n", idxcnt - 1);

    fclose(f1);
    if(mode == 0)
	fclose(f2);

    return(0);
}

/*  Dump abstracts  */
int dumpabstr(int docnt, int mode)
{
    int i, j, k;
    FILE *f1;

    i = 0;
    j = 0;
    k = 0;
    f1 = NULL;

    if (mode == 0) {
	if ((f1 = fopen(CGS_ABSTR_LIST, "w")) == NULL) {
	    fprintf(stderr, "dumpabstr(): Can't open %s\n", CGS_ABSTR_LIST);
	    return(-1);
	}
    } else {
	if ((f1 = fopen(CGS_ABSTR_TXT, "w")) == NULL) {
	    fprintf(stderr, "dumpabstr(): Can't open %s\n", CGS_ABSTR_TXT);
	    return(-1);
	}
    }

    i = 0;
    while (i < docnt) {
	if (mode != 0) {
	    if (lrgflg == 0)
		fprintf(f1, "%04X ", abstrlist[i].words[0]);
	    else
		fprintf(f1, "%08X ", abstrlist[i].words[0]);
	}
	/*  Renumber  */
	j = 1;
	while (j < CGS_ABSTR_SIZ) {
	    k = abstrlist[i].words[j];
	    if(k == 0)
		break;
	    if (compflg == 0)
		abstrlist[i].words[j] = wordlist[k - 1].num;
	    else
		abstrlist[i].words[j] = wrdlstidx[k - 1].num;
	    if (mode != 0) {
		if (lrgflg == 0)
		    fprintf(f1, "%04X", abstrlist[i].words[j]);
		else
		    fprintf(f1, "%08X", abstrlist[i].words[j]);
	    }
	    j++;
	    if (mode != 0) {
		if(abstrlist[i].words[j] != 0)
		    fprintf(f1, " ");
		else
		    fprintf(f1, "\n");
	    }
	}
	abstrlist[i].words[CGS_ABSTR_MAX] = 0;	/*  Terminate  */
	if (mode == 0)
	    fwrite(&abstrlist[i], sizeof(struct abstr), 1, f1);
	i++;
    }

    printf("dumpabstr(): Wrote %d abstracts to abstrlist\n", i);

    fclose(f1);

    return(0);
}

#ifdef CGS_DO_TEST
/*  Dump word pointers for debug  */
int dumpwrdptrs(int mode)
{
    int cnt, i, k;
    FILE *f1;

    i = 0;
    k = 0;
    f1 = NULL;

    cnt = 1;

    if ((f1 = fopen("wrdptrs.dump", "w")) == NULL) {
	fprintf(stderr, "dumpwrdptrs(): Can't open wrdptrs.dump\n");
	return(-1);
    }

    while (i < wordlstuse) {
	if (compflg == 0)
	    k = wordlist[wrdptrs[i]].num;
	else
	    k = wrdlstidx[wrdptrs[i]].num;
	if (docidx[k - 1].next == 0) {
	    i++;
	    continue;
	}
	if (mode == 0)
	    fprintf(f1, "%04X %04X\n", cnt, k);
	else
	    fprintf(f1, "%08X %08X\n", cnt, k);
	cnt++;
	i++;
    }

    fclose(f1);

    return(0);
}
#endif

/*  Count words for statistices  */
void countwords(void)
{
    int i;

    printf("\nCount  Word\n\n");
    for(i = 0; i < wordlstuse; i++)
	printf("%5d  %s\n", docidx[wrdptrs[i]].next, \
	  wordlist[wrdptrs[i]].str);
    printf("\n");
}

/*  Initialize arrays  */
int init(void)
{
    int docidxsiz, i, len;
    char *word;
    FILE *refidxfil, *refwlfil;
    struct stridx index;
    struct wrdentr wordent;

    docidxsiz = 0;
    i = 0;
    len = 0;
    refidxfil = NULL;	/*  Reference index file  */
    refwlfil = NULL;	/*  Reference wordlist file  */
    word = NULL;
    memset(&index, 0, sizeof(index));
    memset(&wordent, 0, sizeof(wordent));

    /*
    *  Intial array sizes
    */

    /*  Wordlist  */
    wordlstentr = CGS_DFLT_WORD_ENTRS;
    wordlstmax  = wordlstentr - 1;
    if (compflg == 0) {
	wordlstsiz  = wordlstentr * sizeof(struct wrdentr);
    } else {
	/*  Compact wordlist  */
	wrdidxsiz  = wordlstentr * sizeof(struct stridx);
	/*  Average of min and max  */
	wordlstsiz = wordlstentr * ((CGS_MIN_WRD_LEN + CGS_WRD_MAX) / 2);
    }
    wordptrsiz  = wordlstentr * sizeof(int);

    /*
    *  initial allocs
    */
    /*  Wordlist  */
    if (compflg == 0) {
	wordlist = malloc((size_t) wordlstsiz);
	if (wordlist == NULL) {
	    fprintf(stderr, "init(): Can't allocate memory for wordlist\n");
	    exit(1);
	}
	memset(wordlist, 0, wordlstsiz);
	if (debflg != 0) {
	    fprintf(stderr, \
	      "Debug: init(): Allocated %d bytes for wordlist at %lx\n", \
	      wordlstsiz, (long unsigned int) wordlist);
	}
    } else {
	/*  Compact wordlist  */
	wrdlstidx = malloc((size_t) wrdidxsiz);
	if (wrdlstidx  == NULL) {
	    fprintf(stderr, "init(): Can't allocate memory for wordlist index\n");
	    exit(1);
	}
	memset(wrdlstidx, 0, wrdidxsiz);
	if (debflg != 0) {
	    fprintf(stderr, \
	      "Debug: init(): Allocated %d bytes for wordlist index at %lx\n", \
	      wrdidxsiz, (long unsigned int) wrdlstidx);
	}
	compwordlst = (char *) malloc((size_t) wordlstsiz);
	if (compwordlst == NULL) {
	    fprintf(stderr, "init(): Can't allocate memory for wordlist\n");
	    free(wrdlstidx);
	    exit(1);
	}
	memset(compwordlst, 0, wordlstsiz);
	if (debflg != 0) {
	    fprintf(stderr, \
	      "Debug: init(): Allocated %d bytes for wordlist at %lx\n", \
	      wordlstsiz, (long unsigned int) compwordlst);
	}
    }
    /*  Word pointers  */
    wrdptrs = malloc((size_t) wordptrsiz);
    if (wrdptrs == NULL) {
	fprintf(stderr, "init(): Can't allocate memory for wrdptrs\n");
	free(wordlist);
	exit(1);
    }
    /*  Set pointers to max. This should never be reached  */
    //for(i = 0; i < wordlstentr; i++)
	//wrdptrs[i] = wordlstmax;
    if (debflg != 0) {
	fprintf(stderr, \
	  "Debug: init(): Allocated %d bytes for wrdptrs at %lx\n", \
	  wordptrsiz, (long unsigned int) wrdptrs);
    }

    i = 0;
    if (reflg != 0 && (refwlfil = fopen(CGS_REF_WRDS_LIST, "r")) != NULL) {
	/*  Use old wordlist as reference  */
	if(debflg != 0)
	    fprintf(stderr, "Debug: init(): opened %s\n", CGS_REF_WRDS_LIST);
	if ((refidxfil = fopen(CGS_REF_WRDS_IDX, "r")) != NULL) {
	    /*  Compact file format  */
	    if(debflg != 0)
		fprintf(stderr, "Debug: init(): opened %s\n", CGS_REF_WRDS_IDX);
        }
	while (1) {
	    /*  Read reference wordlist  */
	    if (refidxfil == NULL) {
		/*  No compact file format in  */
		if (fread(&wordent, sizeof(struct wrdentr), 1, refwlfil) == 0) {
		    if (debflg != 0) {
			fprintf(stderr, \
			  "Debug: init(): Stopped reading file %s\n", \
			  CGS_REF_WRDS_LIST);
		    }
		    break;
		}
		/*  Copy word  */
		if (compflg == 0) {
		    strncpy(wordlist[i].str, wordent.str, CGS_WRD_MAX);
		} else {
		    /*  Compact file format out  */
		    wrdlstidx[i].len = strlen(wordent.str);
		    /*  Limit length  */
		    if(wrdlstidx[i].len > CGS_WRD_MAX)
			wrdlstidx[i].len = CGS_WRD_MAX; /*  Limit size  */
		    if (i == 0) {
			/*  1st entry  */
			wrdlstidx[0].ptr = 0;
			word = compwordlst;
		    } else {
			wrdlstidx[i].ptr = \
			  wrdlstidx[i - 1].ptr + wrdlstidx[i - 1].len + 1;
			word = compwordlst + wrdlstidx[i].ptr;
		    }
		    strncpy(word, wordent.str, wrdlstidx[i].len);
		    /*  Terminate  */
		    word[wrdlstidx[i].len] = 0;
		}
	    } else {
		/*  Compact file format in  */
		if (fread(&index, sizeof(struct stridx), 1, refidxfil) == 0) {
		    if (debflg != 0) {
			fprintf(stderr, \
			  "Debug: init(): Stopped reading file %s\n", \
			  CGS_REF_WRDS_IDX);
		    }
		    break;
		}
		/*  Copy word  */
		len = index.len;
		if(len > CGS_WRD_MAX)
		    len = CGS_WRD_MAX;  /*  Limit size  */
		if (compflg == 0) {
		    word = wordlist[i].str;
		} else {
		    /*  Compact file format out  */
		    if (i == 0) {
			wrdlstidx[0].ptr = 0;
			word = compwordlst;
		    } else {
			wrdlstidx[i].ptr = wrdlstidx[i - 1].ptr + wrdlstidx[i - 1].len + 1;
			word = compwordlst + wrdlstidx[i].ptr;
		    }
		}
		if (fread(word, 1, len + 1, refwlfil) != len + 1) {
		    if (debflg != 0) {
			fprintf(stderr, \
			  "Debug: init(): Stopped reading file %s\n", \
			  CGS_REF_WRDS_LIST);
		    }
		    break;
		}
		/*  Terminate  */
		word[len] = 0;
		if (compflg != 0) {
		    /*  Update index  */
		    wrdlstidx[i].len = len;
		}
		if (len != index.len) {
		    fprintf(stderr, \
		      "init(): Warning: Length mismatch at %04X\n", i);
		    fseek(refwlfil, (long) (index.len - len), SEEK_CUR);
		}
	    }
	    /*  And renumber  */
	    if (compflg == 0)
		wordlist[i].num = i + 1;        /*  Starts numbering at 1  */
	    else
		wrdlstidx[i].num = i + 1;
	    i++;
	    /*  Check wordlstuse size and realloc if needed  */
	    if (i > wordlstmax) {
		if (lrgflg == 0 && wordlstentr > 65530) {
		    /*  16 bit word numbers  */
		    fprintf(stderr, "init(): Number of words too large\n");
		    freethelot(1);
		}
		/*  Increase size of various tables  */
		growtables(i);
	    }
	    /*  Separate check for compact worldlist  */
	    if (compflg != 0) {
		/*
		*  Foobar0CGS_WRD_SI0
		*  ^     ^^ ptr 
		*  ptr-- len--
		*/
		if (wrdlstidx[i - 1].ptr + wrdlstidx[i - 1].len + \
		  CGS_WRD_SIZ >= wordlstsiz) {
		    growcompwrdlst();
		}
	    }
	}
	if (refidxfil != NULL)
	    fclose(refidxfil);
 	fclose(refwlfil);
	if (debflg != 0) {
	    fprintf(stderr, "Debug: init(): read %d words from %s\n", i, \
	      CGS_REF_WRDS_LIST);
	}
    } else {
	/*  The first few default entries  */
	while (defltlst[i] != NULL && defltlst[i][0] != 0) {
	    if (compflg == 0) {
		strcpy(wordlist[i].str, defltlst[i]);
		wordlist[i].num = i + 1;
	    } else {
		/*  Compact wordlist  */
		wrdlstidx[i].num = i + 1;
		wrdlstidx[i].len = strlen(defltlst[i]);
		strcpy(compwordlst + wrdlstidx[i].ptr, defltlst[i]);
		/*  Next pointer; There should always be a margein of one!  */
		wrdlstidx[i + 1].ptr = wrdlstidx[i].ptr + wrdlstidx[i].len + 1;
	    }
	    i++;
	}
    }
    wordlstuse = i;	/*  Points to next free pos in array  */

    /*  Set the pointers  */
    for(i = 0; i < wordlstuse; i++)
       wrdptrs[i] = i;

    /*  Word to doc index initial alloc  */
    docidxsiz = wordlstentr * sizeof(struct word2docs);
    docidx = malloc((size_t) docidxsiz);
    if (docidx == NULL) {
	fprintf(stderr, "init(): Can't allocate memory for docidx.\n");
	free(wordlist);
	free(wrdptrs);
	exit(1);
    }
    memset(docidx, 0, docidxsiz);
    for (i = 0; i < wordlstentr; i++) {
	docidx[i].docs = &(docidx[i].data[0]);  /*  Points to data  */
	docidx[i].siz = (unsigned short) CGS_DFLT_DAT_SIZ;
    }
    if (debflg != 0) {
	fprintf(stderr, \
	  "Debug: init(): Allocated %d bytes for docidx at %lx\n", \
	  docidxsiz, (long unsigned int) docidx);
    }

    if (absflg != 0) {
	/*  Abstracts initial alloc */
	abstrentr = CGS_DFLT_DOCS;
	abstrmax = abstrentr - 1;
	abstrsiz = abstrentr * sizeof(struct abstr);
	abstrlist = malloc((size_t) abstrsiz);
	if (abstrlist == NULL) {
	    fprintf(stderr, "init(): Can't allocate memory for abstrlist\n");
	    free(docidx);
	    free(wrdptrs);
	    free(wordlist);
	    exit(1);
	}
	memset(abstrlist, 0, abstrsiz);
	if (debflg != 0) {
	    fprintf(stderr, \
	      "Debug: init(): Allocated %d bytes for abstrlist at %lx\n", \
	      abstrsiz, (long unsigned int) abstrlist);
	}
    }

    #ifdef CGS_WITH_HTTP
     /*  URL list  */
     urlstentr = CGS_DFLT_DOCS;
     urlstmax = urlstentr - 2;	/*  Extra margin  */
     urlstsiz = urlstentr * sizeof(struct urlentr);
     urlist = malloc((size_t) urlstsiz);
     if (urlist == NULL) {
	 fprintf(stderr, "init(): Can't allocate memory for urlst\n");
	 if(abstrlist != NULL)
	     free(abstrlist);
	 free(docidx);
	 free(wrdptrs);
	 free(wordlist);
	 exit(1);
     }
     memset(urlist, 0, urlstsiz);
     if (debflg != 0) {
	 fprintf(stderr, \
	   "Debug: init(): Allocated %d bytes for urlist at %lx\n", \
	   urlstsiz, (long unsigned int) urlist);
     }

     /*  URL pointers has fixed size */
     memset(urlptrs, 0, sizeof(urlptrs));
     if (debflg != 0) {
	 fprintf(stderr, \
	   "Debug: init(): urlptrs has fixed size of 262144 bytes\n");
     }

     /*  External URL list  */
     if (extnflg != 0) {
	 /*  External links report  */
	 extlstentr = CGS_DFLT_DOCS;
	 extlstmax  = extlstentr - 1;
	 extlstsiz  = extlstentr * sizeof(struct exturl);
	 extnlst = malloc((size_t) extlstsiz);
	 if (extnlst == NULL) {
	     fprintf(stderr, "init(): Can't allocate memory for extnlst\n");
	     if(abstrlist != NULL)
		 free(abstrlist);
	     free(docidx);
	     free(wrdptrs);
	     free(wordlist);
	     free(urlist);
	     exit(1);
	 }
	 memset(extnlst, 0, extlstsiz);
	 if (debflg != 0) {
	     fprintf(stderr, \
	       "Debug: init(): Allocated %d bytes for extnlst at %lx\n", \
	       extlstsiz, (long unsigned int) extnlst);
	     fprintf(stderr, \
	       "Debug: init(): extptrs has fixed size of 262144 bytes\n");
	 }
     }
     memset(extptrs, 0, sizeof(extptrs));

     //curlh = NULL;	Moved to main()
     /* curl_global_init(long flags); */
     if (curl_global_init(CURL_GLOBAL_SSL) != 0) {
	 fprintf(stderr, "init(): curl_global_init() failed\n");
	 free(urlist);
	 if(abstrlist != NULL)
	     free(abstrlist);
	 free(docidx);
	 free(wrdptrs);
	 free(wordlist);
	 exit(1);
     }
     if ((curlh = curl_easy_init()) == NULL) {
	 fprintf(stderr, "init(): curl_easy_init() failed\n");
	 curl_global_cleanup();
	 free(urlist);
	 if(abstrlist != NULL)
	     free(abstrlist);
	 free(docidx);
	 free(wrdptrs);
	 free(wordlist);
	 exit(1);
     }
    #endif

    return(0);
}

/*  Grow abstracts  */
int growabstr(void)
{
    int size;

    size = 2 * abstrsiz;

    abstrlist = (struct abstr *) realloc(abstrlist, size);
    if (abstrlist == NULL) {
	fprintf(stderr, "growabstr(): Can't allocate memory for abstrlist\n");
	freethelot(1);
    }
    memset(abstrlist + (abstrsiz / sizeof(struct abstr)), 0, abstrsiz);
    if (debflg != 0) {
	fprintf(stderr, \
	  "Debug: growabstr(): Growing abstrlist from %d to %d bytes at %lx\n", \
	  abstrsiz, size, (unsigned long) abstrlist);
    }

    abstrsiz = size;
    abstrentr = abstrentr * 2;
    abstrmax = abstrentr - 1;

    return(0);
}

/*  Initialize synonyms  */
int initsyn(void)
{
    int i, size, synidxfd, synidxsiz;
    char *synlstab;
    struct stat sb;
    struct synidx idx;

    i = 0;
    size = 0;
    synidxfd = 0;	/*  Synonym index FD  */
    synidxsiz = 0;	/*  Synonym index size  */
    synlstab = NULL;	/*  Char version of synlstadr  */
    memset(&idx, 0, sizeof(idx));
    memset(&sb, 0, sizeof(sb));

    if ((synidxfd = open(CGS_SYNOM_IDX, O_RDONLY)) < 0) {
	fprintf(stderr, "initsyn(): Cant open %s\n", CGS_SYNOM_IDX);
	return(-1);
    }

    if ((synlstfd = open(CGS_SYNOM_LIST, O_RDONLY)) < 0) {
	fprintf(stderr, "initsyn(): Cant open %s\n", CGS_SYNOM_LIST);
	synlstfd = 0;
	close(synidxfd);
	return(-1);
    }

    /*  Index size  */
    if (fstat(synidxfd, &sb) != 0) {
	fprintf(stderr, "initsyn(): fstat failed\n");
	close(synlstfd);
	synlstfd = 0;
	close(synidxfd);
	return(-1);
    }
    synidxsiz = sb.st_size;
    synentr = synidxsiz / sizeof(struct synidx);
    size = synentr * sizeof(struct synonym);

    /*  Synonyms table  */
    syntab = malloc(size);
    if (syntab == NULL) {
	fprintf(stderr, "initsyn(): Can't allocate memory for syntab\n");
	close(synlstfd);
	synlstfd = 0;
	close(synidxfd);
	return(-1);
    }
    memset(syntab, 0, size);
    if (debflg != 0) {
	fprintf(stderr, \
	  "Debug: initsyn(): Allocated %d bytes for syntab at %lx\n", \
	  size, (long unsigned int) syntab);
    }

    /*  Map list  */
    if (fstat(synlstfd, &sb) != 0) {
	fprintf(stderr, "initsyn(): fstat failed\n");
	free(syntab);
	syntab = NULL;
	close(synlstfd);
	synlstfd = 0;
	close(synidxfd);
	return(-1);
    }
    synlstsiz = (size_t) sb.st_size;
    synlstadr = mmap(NULL, synlstsiz, PROT_READ, MAP_PRIVATE, synlstfd, 0);
    if (synlstadr == MAP_FAILED) {
	fprintf(stderr, "initsyn(): mmap failed\n");
	free(syntab);
	syntab = NULL;
	close(synlstfd);
	synlstfd = 0;
	close(synidxfd);
	return(-1);
    }
    /*  Cast synonyms to char  */
    synlstab = (char *) synlstadr;

    /*  Fill table  */
    i = 0;
    while (i < synentr) {
	if (read(synidxfd, &idx, sizeof(idx)) != sizeof(idx)) {
	    fprintf(stderr, "initsyn(): Error reading %s\n", CGS_SYNOM_IDX);
	    free(syntab);
	    syntab = NULL;
	    munmap(synlstadr, synlstsiz);
	    synlstadr = NULL;
	    close(synlstfd);
	    synlstfd = 0;
	    close(synidxfd);
	    return(-1);
	}
	syntab[i].org = synlstab + idx.org;
	syntab[i].syn = synlstab + idx.syn;
	i++;
    }
    if (debflg != 0) {
        fprintf(stderr, \
          "Debug: initsyn(): %d entries in syntab\n", i);
    }

    /*  Done with index file  */
    close(synidxfd);

    return(0);
}

/*  Process file list  */
int procfilst(char *listn, struct timespec pause)
{
    int chrcnt, docnt, htmflg, len, lincnt, ret, tags;
    char wdocn[CGS_PATH_SIZ];
    char line[CGS_BUF_SIZ], line2[CGS_BUF_SIZ];
    struct timespec rem;
    FILE *listf, *wdocf;

    chrcnt = 0;				/*  Nr of chars (actually bytes)  */
    docnt  = 0;				/*  Nr of documents  */
    htmflg = 0;				/*  1 if document is HTML  */
    len    = 0;		
    lincnt = 0;				/*  Line count  */
    ret    = 0;		
    tags   = 0;		
    listf = NULL;			/*  List file file  */
    wdocf = NULL;			/*  Web doc file  */
    memset(line,  0, CGS_BUF_SIZ);
    memset(line2, 0, CGS_BUF_SIZ);
    memset(&rem,  0, sizeof(rem));
    memset(wdocn, 0, CGS_PATH_SIZ);	/*  Web doc name  */

    if ((listf = fopen(listn, "r")) == NULL) {
	fprintf(stderr, "procfilst(): Can't open %s\n", listn);
	freethelot(1);
    }

    /*  Read file list file  */
    while (fgets(wdocn, CGS_PATH_SIZ, listf)) {
	len = strlen(wdocn) - 1;
	if (wdocn[len] == 10) {
	    wdocn[len] = 0;
	    len--;
	}
	if(wdocn[0] == 0)
	    continue;
	if (wdocn[len] == 13) {
	    wdocn[len] = 0;
	    len--;
	}
	if(wdocn[0] == 0)
	    continue;
	len++;
	#ifdef CGS_MULT_HTML_EXT
	 htmflg = 0;
	 i = 0;
	 while (htmlexts[i].len != 0) {
	     if (len > htmlexts[i].len && \
	       strcmp(wdocn + len - htmlexts[i].len, htmlexts[i].str) == 0) {
		 htmflg = 1;
		 break;
	     }
	     i++;
	 }
	#else
	 if (len > 5 && strcmp(wdocn + len - 5, ".html") == 0)
	     htmflg = 1;
	 else
	     htmflg = 0;
	#endif

	if ((wdocf = fopen(wdocn, "r")) == NULL) {
	    fprintf(stderr, "procfilst(): Can't open %s\n", wdocn);
	    fclose(listf);
	    freethelot(1);
	}

	if (debflg != 0) {
	    if (htmflg == 0)
		fprintf(stderr, "Debug: procfilst(): Opened Text file %s\n", wdocn);
	    else
		fprintf(stderr, "Debug: procfilst(): Opened HTML file %s\n", wdocn);
	}

	lincnt = 1;	/*  Line count  */
	tags = 0;	/*  Reset tags  */
	while (fgets(line, CGS_BUF_SIZ, wdocf)) {
	    len = strlen(line);
	    ret = 0;
	    if (htmflg != 0) {
		#ifdef CGS_WITH_HTTP
		 /* int remtag(char *in, char *out, int len, char *mhost, \
		   char *base, int *tags, struct titlprps *title, \
		   int docnt, int crawl) */
		 ret = remtag(line, line2, len, NULL, NULL, &tags, NULL, -1, 0);
		#else
		 ret = remtag(line, line2, len, &tags);
		#endif
		if (ret < 0) {
		    fprintf(stderr, \
		      "procfilst(): Indexing of %s terminated at line %d\n", \
		      wdocn, lincnt);
		    break;
		} else if (ret > 0) {
		    ret = findwords(line2, ret, docnt, &chrcnt);
		}
	    } else {
		ret = findwords(line, len, docnt, &chrcnt);
	    }
	    if (ret != 0) {
		fprintf(stderr, "procfilst(): %s is not a UTF-8 file\n", wdocn);
		fprintf(stderr, \
		  "procfilst(): Indexing of %s terminated at line %d\n", \
		  wdocn, lincnt);
		break;
	    }
	    lincnt++;
	}

	fclose(wdocf);
	docnt++;
	/*  Check maximum  */
	if (docnt > 65530) {
	    fprintf(stderr, "procfilst(): Too many documents.\n");
	    freethelot(1);
	}
	/*  Check abstracts  */
	if(absflg != 0 && docnt > abstrmax)
	    growabstr();
	/*  Wait, if any  */
	if (pause.tv_sec != 0 || pause.tv_nsec != 0) {
	    if (nanosleep(&pause, &rem) != 0) {
		fprintf(stderr, "procfilst(): nanosleep() error\n");
		if(errno == EINTR)
		    nanosleep(&rem, &rem);
	    }
	}

    }	/*  End of read file list file  */
    /*  Close  */
    fclose(listf);

    printf("procfilst(): Found %d words in %d documents\n", \
      wordlstuse, docnt);
    printf("procfilst(): Max wordlength is %d bytes\n", chrcnt);

    return(docnt);
}

/*  End of generic bits  */

/*  START of HTTP specific bits  */
#ifdef CGS_WITH_HTTP

/*  Convert to UTF-8  */
int toutf(char *inbuf, int inlen, char *outbuf, int outsiz, char *charset)
{
    /*  fromcode at least two larger then CGS_CHARSET_SIZ  */
    char fromcode[32], *ip, *op;
    int outlen;
    size_t inbytesleft, outbytesleft;
    iconv_t cd;

    outlen = 0;
    memset(fromcode, 0, 32);

    inbytesleft  = (size_t) inlen;
    outbytesleft = (size_t) outsiz;
    ip = inbuf;
    op = outbuf;

    snprintf(fromcode, 31, "%s//", charset);
    if (debflg != 0) {
	fprintf(stderr, "Debug: toutf(): len %d fromcode: %s\n", \
	  inlen, fromcode);
    }

    if ((cd = iconv_open("UTF-8//", fromcode)) < 0) {
	fprintf(stderr, "toutf(): iconv_open failed\n");
	return(-1);
    }
    if (iconv(cd, &ip, &inbytesleft, &op, &outbytesleft) < 0) {
	fprintf(stderr, "toutf(): iconv failed\n");
	iconv_close(cd);
	return(-1);
    }
    iconv_close(cd);

    outlen = outsiz - ((int) outbytesleft);
    outbuf[outlen] = 0;		/*  Terminate  */
    if (debflg != 0) {
	fprintf(stderr, "Debug: toutf(): converted to len: %d\n", \
	  outlen);
    }

    /*  Return stringlength  */
    return(outlen);
}

/*  Convert from UTF-8  */
int fromutf(char *inbuf, int inlen, char *outbuf, int outsiz, char *charset)
{
    /*  tocode at least two larger then CGS_CHARSET_SIZ  */
    char tocode[32], *ip, *op;
    int outlen;
    size_t inbytesleft, outbytesleft;
    iconv_t cd;

    outlen = 0;
    memset(tocode, 0, 32);

    inbytesleft  = (size_t) inlen;
    outbytesleft = (size_t) outsiz;
    ip = inbuf;
    op = outbuf;

    snprintf(tocode, 31, "%s//", charset);
    if (debflg != 0) {
	fprintf(stderr, "Debug: fromutf(): len %d tocode: %s\n", \
	  inlen, tocode);
    }
    if ((cd = iconv_open(tocode, "UTF-8//")) < 0) {
	fprintf(stderr, "fromutf(): iconv_open failed\n");
	return(-1);
    }
    if (iconv(cd, &ip, &inbytesleft, &op, &outbytesleft) < 0) {
	fprintf(stderr, "fromutf(): iconv failed\n");
	iconv_close(cd);
	return(-1);
    }
    iconv_close(cd);

    outlen = outsiz - ((int) outbytesleft);
    outbuf[outlen] = 0;		/*  Terminate  */
    if (debflg != 0) {
	fprintf(stderr, "Debug: fromutf(): converted to len: %d\n", \
	  outlen);
    }

    /*  Return stringlength  */
    return(outlen);
}

/*  Add URL to URL list  */
int addurl(char *url, int compurlstat)
{
    int i, j, ret;

    i = 0;
    j = 0;
    ret = 0;

    /*
    *  urlist[x]		Order is added urls
    *  urlist[wrdptrs[x]]	Order is alphabetical
    */

    /*  compurlstat is status of urlsrch()  */
    i = compurlstat;

    if(i < 0)
        i = 0;

    while (i < urlstuse) {
	if (strcmp(url, urlist[urlptrs[i]].url) < 0) {
	    /* Str < urlist[i].url  */
	    j = i;      /*  To be pos of URL  */
	    break;
	}
	i++;
    }

    if (i == urlstuse) {
	/*  New entry all the way at the back  */
	strcpy(urlist[urlstuse].url, url);
	urlist[urlstuse].num = (unsigned short) (urlstuse + 1);
	urlptrs[urlstuse] = urlstuse;
	j = urlstuse; /*  For debug  */
    } else {
	i = urlstuse;
	while (i > j) {
	    urlptrs[i] = urlptrs[i - 1];
	    i--;
	}
	/*  New entry  */
	strcpy(urlist[urlstuse].url, url);
	urlist[urlstuse].num = (unsigned short) (urlstuse + 1);
	urlptrs[j] = urlstuse;
    }
    if (debflg != 0) {
	fprintf(stderr, "Debug: addurl(): Added Url: %04X %s\n", \
	  urlist[urlstuse].num, urlist[urlstuse].url);
    }

    ret = urlstuse;
    urlstuse++;
    if (urlstuse > 65530) {
	fprintf(stderr, "addurl(): Too many documents\n");
	freethelot(1);
    }

    /*  Check urlstuse size and realloc if needed  */
    if (urlstuse > urlstmax) {
	/*  Increase size of urlist  */
	growurlst();
    }

    /*  Return number of new URL  */
    return(ret);
}

/*  Add external URL to external URL list  */
int addextn(char *url, int compurlstat)
{
    int i, j, ret;

    i = 0;
    j = 0;
    ret = 0;

    /*
    *  extnlst[x]		Order is added urls
    *  extnlst[extptrs[x]]	Order is alphabetical
    */

    /*  compurlstat is status of extnsrch()  */
    i = compurlstat;

    if(i < 0)
	i = 0;

    while (i < extlstuse) {
	if (strcmp(url, extnlst[extptrs[i]].url) < 0) {
	    /* Str < extnlst[i].url  */
	    j = i;
	    break;
	}
	i++;
    }

    if (i == extlstuse) {
	/*  New entry all the way at the back  */
	strcpy(extnlst[extlstuse].url, url);
	extptrs[extlstuse] = extlstuse;
	j = extlstuse; /*  For debug  */
    } else {
	i = extlstuse;
	while (i > j) {
	    extptrs[i] = extptrs[i - 1];
	    i--;
	}
	/*  New entry  */
	strcpy(extnlst[extlstuse].url, url);
	extptrs[j] = extlstuse;
    }

    if (debflg != 0) {
	fprintf(stderr, "Debug: addextn(): Added Url: %04X %s\n", \
 	  extlstuse + 1, extnlst[extlstuse].url);
    }

    ret = extlstuse;
    extlstuse++;
    if (extlstuse > 65530) {
	fprintf(stderr, "addextn(): Too many documents\n");
	freethelot(1);
    }

    /*  Check extlstuse size and realloc if needed  */
    if (extlstuse > extlstmax) {
	/*  Increase size of extnlst  */
	growextnlst();
    }

    /*  Return number of new URL  */
    return(ret);
}

/*  Search URL  */
int urlsrch(char *key, int *compurlstat)
{
    int first, last, middle, ret;

    first = 0;
    ret = 0;

    last = urlstuse - 1;
    middle = (first + last) / 2;

    while (first <= last) {
	ret = strcmp(urlist[urlptrs[middle]].url, key);
	if (ret < 0) {
	    /*  str < key. Examine 2nd half  */
	    first = middle + 1;
	} else if (ret == 0) {
	    /*  str == key; Found, Return URL number  */
	    return(urlptrs[middle]);
	} else {
	    /*  str > key. Examine 1st half  */
	    last = middle - 1;
	}
	middle = (first + last) / 2;
    }

    *compurlstat = middle;
    if (ret > 0) {
	/*  Too far  */
	(*compurlstat)--;
    }

    return(-1);
}

/*  Search external URL  */
int extnsrch(char *key, int *compurlstat)
{
    int first, last, middle, ret;

    first = 0;
    ret = 0;

    last = extlstuse - 1;
    middle = (first + last) / 2;

    while (first <= last) {
	ret = strcmp(extnlst[extptrs[middle]].url, key);
	if (ret < 0) {
	    /*  str < key. Examine 2nd half  */
	    first = middle + 1;
	} else if (ret == 0) {
	    /*  str == key; Found, Return URL number  */
	    return(extptrs[middle]);
	} else {
	    /*  str > key. Examine 1st half  */
	    last = middle - 1;
	}
	middle = (first + last) / 2;
    }

    *compurlstat = middle;
    if (ret > 0) {
	/*  Too far  */
	(*compurlstat)--;
    }

    return(-1);
}

/*  Find URL  */
int findurl(char *url, int mode)
{
    int  compurlstat;
    char *key = url;
    int  res, ret;

    compurlstat = 0;
    res = 0;
    ret = 0;

    if (mode == 0) {
	/*  Internal URL  */
	res = urlsrch(key, &compurlstat);
    } else {
	/*  External URL  */
	res = extnsrch(key, &compurlstat);
    }

    if (res < 0) {
	/*  URL not found, add to list  */
	if (debflg != 0)
	    fprintf(stderr, "Debug: findurl(): Adding Url %s\n", url);
	if (mode == 0) {
	    /*  Internal  */
	    ret = addurl(url, compurlstat);
	} else {
	    /*  External  */
	    ret = addextn(url, compurlstat);
	}
    } else {
	ret = res;
    }

    /*  Return URL number  */
    return(ret);
}

/*  Compare ref URLs  */
int compref(const void *r1, const void *r2)
{
    int len;
    struct refurl *ref1, *ref2;

    ref1 = (struct refurl *) r1;
    ref2 = (struct refurl *) r2;

    /*  One of these contains the length  */
    if(ref1->len == 0)
	len = ref2->len;
    else
	len = ref1->len;

    return(strncmp(ref1->url, ref2->url, len));
}

/*  Is URL part of reference list?  */
int chkrefs(char *url, char *mhost)
{
    struct refurl key, *res;

    memset(&key, 0, sizeof(key));
    res = NULL;

    if (strncasecmp(url, mhost, strlen(mhost)) == 0) {
	/*  Matches host  */
	return(0);
    }

    strcpy(key.url, url);
    res = bsearch(&key, refurltab, refurluse, sizeof(struct refurl), compref);
    if (res == NULL) {
	/*  External URL  */
	if(debflg != 0)
	    fprintf(stderr, "Debug: chkrefs(): External url: %s\n", url);
	return(-1);
    }

    return(0);
}

/*  Strip URL  */
int stripurl(char *url, int maxlen)
{
    int i;

    i = 0;
    while (i < maxlen) {
	if (url[i] == 0)
	    break;
	/*  Truncate at '#'  */
	if (url[i] == '#') {
	    /*  Just truncate and hope for the best  */
	    if (debflg != 0) {
		fprintf(stderr, \
		  "Debug: stripurl(): Char 0x%02X, Url truncated\n", url[i]);
	    }
	    url[i] = 0;
            break;
        }
        i++;
    }

    return(i);
}

/*  Construct path  */
int procpath(char *mhost, char *base, char *link)
{
    int dircnt, i, lnkptr;
    char strpath[CGS_PATH_SIZ], fullurl[CGS_BUF_SIZ];

    /*  Returns valid internal URL index number or -1  */

    dircnt = 0;
    i = 0;
    lnkptr = 0;
    memset(strpath, 0, CGS_PATH_SIZ);
    memset(fullurl, 0, CGS_BUF_SIZ);

    if (debflg != 0) {
	fprintf(stderr, "Debug: procpath(): Host: %s Base: %s Link: %s\n", \
	  mhost, base, link);
    }

    if (link[0] == '#' || link[0] == '?') {
	/*  Links to self  */
	return(-1);
    } else if (strncasecmp(link, "http://", 7) == 0 || \
      strncasecmp(link, "https://", 8) == 0 || \
      strncasecmp(link, "ftp://", 6) == 0 || \
      strncasecmp(link, "mailto:", 7) == 0) {
	if (chkrefs(link, mhost) == 0) {
	    /*  Internal link  */
	    strncpy(fullurl, link, CGS_BUF_MAX);
	} else {
	    /*  External link  */
	    if (extnflg != 0) {
		/*  Add to external URL list  */
		stripurl(link, CGS_PATH_SIZ);
		findurl(link, 1);
	    }
	    strcpy(fullurl, "External");
	}
    } else if (strncasecmp(link, "/~", 2) == 0) {
	/*  Public_html  */
	snprintf(fullurl, CGS_BUF_MAX, "%s%s", mhost, link + 1);
    } else if (link[0] == '/') {
	/*  Absolute link  */
	snprintf(fullurl, CGS_BUF_MAX, "%s%s", mhost, link + 1);
    } else if (link[0] == '~') {
	/*  Public_html  */
	snprintf(fullurl, CGS_BUF_MAX, "%s%s", mhost, link);
    } else {
	/* relative link  */
	if (strncmp(link, "../", 3) == 0) {
	    //printf("Debug: B: %s L: %s\n", base, link);
	    dircnt = 1;	/*  This is actually a '/' count  */
	    /*  Find nr of ../  */
	    i = 0;
	    while (i < CGS_PATH_SIZ) {
		if (strncmp(link + i, "../", 3) != 0)
		    break;
	        dircnt++;
	        i += 3;
	    }
	    lnkptr = i;

	    /*  Copy base to strpath and find length  */
	    i = 0;
	    while (i < CGS_PATH_MAX) {
		strpath[i] = base[i];
		if(base[i] == 0)
		    break;
		i++;
	    }

	    while (i > 0) {
		if (strpath[i] == '/') {
		    dircnt--;
		    if(dircnt == 0)
			break;
		}
		i--;
	    }
	    strpath[i + 1] = 0;
	    snprintf(fullurl, CGS_BUF_MAX, "%s%s", strpath, link + lnkptr);
	} else if (link[0] == '.') {
	    /*  To current dir  */
	    snprintf(fullurl, CGS_BUF_MAX, "%s", base);
	} else {
	    snprintf(fullurl, CGS_BUF_MAX, "%s%s", base, link + lnkptr);
	}
    }

    /*  Strip URL  */
    if (stripurl(fullurl, CGS_BUF_SIZ) > CGS_PATH_MAX2) {
	fprintf(stderr, "procpath(): Url too long\n");
	return(-1);
    }

    if(debflg != 0)
	fprintf(stderr, "Debug: procpath(): Url: %s\n", fullurl);

    if (strcmp(fullurl, "External") == 0) {
	/* External URL  */
	return(-1);
    } else {
	/* Internal URL  */
	return(findurl(fullurl, 0));
    }
}

/*  Find base URL  */
int fndbasurl(char *url, char *base, char *mhost)
{
    int i, j, slashcnt;

    i = 0;
    j = 0;
    slashcnt = 0;
    while (i < CGS_PATH_MAX) {
	base[i] = url[i];
	if (slashcnt < 3) {
	    mhost[j] = url[i];
	    j++;
	    /*  Method://Host/ */
	    if(url[i] == '/')
		slashcnt++;
	}
	if(url[i] == 0)
	    break;
	i++;
    }
    base[i] = 0;
    mhost[j] = 0;
    while (i > 0) {
	if (base[i] == '/') {
	    base[i + 1] = 0;
	    break;
	}
	i--;
    }

    if (debflg != 0) {
	fprintf(stderr, "Debug: fndbasurl(): Host: %s Base: %s Url: %s\n", \
	  mhost, base, url);
    }

    return(i + 1);
}

/*  Grow URL List */
int growurlst(void)
{
    int size;

    size = 2 * urlstsiz;

    urlist = (struct urlentr *) realloc(urlist, size);
    if (urlist == NULL) {
        fprintf(stderr, "growurlst(): Can't allocate memory for urlist\n");
        freethelot(1);
    }
    memset(urlist + urlstentr, 0, urlstsiz);
    if (debflg != 0) {
	fprintf(stderr, \
	  "Debug: growurlst(): Growing urlist from %d to %d bytes at %lx\n", \
	  urlstsiz, size, (long unsigned int) urlist);
    }
    urlstentr = urlstentr * 2;
    urlstmax = urlstentr - 2;	/*  Extra margin  */
    urlstsiz = size;

    /*  Note: urlptrs has fixed size  */
    return(0);
}

/*  Grow external URL list  */
int growextnlst(void)
{
    int size;

    size = 2 * extlstsiz;

    extnlst = (struct exturl *) realloc(extnlst, size);
    if (extnlst == NULL) {
	fprintf(stderr, "growextnlst(): Can't allocate memory for extnlst\n");
	freethelot(1);
    }
    memset(extnlst + extlstentr, 0, extlstsiz);
    if (debflg != 0) {
	fprintf(stderr, \
	  "Debug: growextnlst(): Growing external urlist from %d to %d bytes at %lx\n", \
	  extlstsiz, size, (long unsigned int) extnlst);
    }
    extlstentr = extlstentr * 2;  
    extlstmax = extlstentr - 1;
    extlstsiz = size;

    return(0);
}

/*  Dump URL list  */
int dumpurlst(int docnt, int mode)
{
    int i, len;
    char urltitl[CGS_BUF_SIZ];
    struct stridx urlindex;
    FILE *fout, *fidx;

    i = 0;
    len = 0;
    memset(urltitl, 0, CGS_BUF_SIZ);
    memset(&urlindex, 0, sizeof(urlindex));
    fout = NULL;
    fidx = NULL;

    if (mode == 0) {
	/*  links-list  */
	if ((fout = fopen(CGS_LNKSLIST, "w")) == NULL) {
	   fprintf(stderr, "dumpurlst(): Can't open %s\n", CGS_LNKSLIST);
	   return(-1);
	}
	/*  links.idx  */
	if ((fidx = fopen(CGS_LNKS_IDX, "w")) == NULL) {
	   fprintf(stderr, "dumpurlst(): Can't open %s\n", CGS_LNKS_IDX);
	   fclose(fout);
	   return(-1);
	}
    } else {
	/*  num-links.list  */
	if ((fout = fopen(CGS_LINKS_TXT, "w")) == NULL) {
	    fprintf(stderr, "dumpurlst(): Can't open %s\n", CGS_LINKS_TXT);
	    return(-1);
	}
    }

    while (i < docnt) {
	if (urlist[i].redir != 0 || urlist[i].stat != 200 || \
	  (urlist[i].type > CGS_TYP_HTML && \
	  (pdflg == 0 || urlist[i].type != CGS_TYP_PDF))) {
	    /*  For debug: This entry should never be displayed  */
	    snprintf(urltitl, CGS_BUF_MAX, \
	      "<a href=\"%s\">%04X %04X %03d %s</a>", \
	      urlist[i].url, (unsigned int) urlist[i].num, \
	      (unsigned int) urlist[i].redir, (int) urlist[i].stat, \
	      typtxts[urlist[i].type]);
	} else if (urlist[i].title[0] == 0) {
	    /*  No title: Use URL  */
	    snprintf(urltitl, CGS_BUF_MAX, "<a href=\"%s\">%s</a>", \
	      urlist[i].url, urlist[i].url);
	} else {
	    snprintf(urltitl, CGS_BUF_MAX, "<a href=\"%s\">%s</a>", \
	      urlist[i].url, urlist[i].title);
	}

	if (mode == 0) {
	    urlindex.num = (int) urlist[i].num;
	    len = strlen(urltitl);
	    urlindex.len = len;
	    fwrite(urltitl, 1, len + 1, fout);
	    fwrite(&urlindex, sizeof(urlindex), 1, fidx);
	    urlindex.ptr = urlindex.ptr + len + 1;	/*  Next  */
	} else {
	    fprintf(fout, "%04X %s\n", (unsigned int) urlist[i].num, urltitl);
	}

	i++;
    }

    fclose(fout);
    if(mode == 0)
	fclose(fidx);

    printf("dumpurlst(): Wrote %d urls to urlist\n", docnt);

    return(0);
}

/*  Report internal and external URLs  */
int reporturls(void)
{
    int i, j;
    FILE *fout;

    i = 0;
    j = 0;
    fout = NULL;

    if ((fout = fopen(CGS_INT_URL_RP, "w")) == NULL) {
	fprintf(stderr, "Can't open %s\n", CGS_INT_URL_RP);
	return(-1);
    }

    fprintf(fout, "Number\tRedir\tStatus\tType\tCharset\tUrl\n\n");
    while (i < urlstuse) {
	j = urlptrs[i];		/*  Alphabetical order  */
	fprintf(fout, "%04X\t%04X\t%03d\t%s\t", urlist[j].num, \
	  urlist[j].redir, urlist[j].stat, typtxts[urlist[j].type]);
	if (urlist[j].charset[0] == 0)
	    fprintf(fout, "Default");
	else
	    fprintf(fout, "%s", urlist[j].charset);
	fprintf(fout, "\t%s\n", urlist[j].url);
	i++;
    }

    fclose(fout);

    printf("reporturls(): Wrote %d urls to internal urls report\n", i);

    if (extnflg != 0) {
	/*  Print external link report  */
	if ((fout = fopen(CGS_EXT_URL_RP, "w")) == NULL) {
	    fprintf(stderr, "Can't open %s\n", CGS_EXT_URL_RP);
	    return(-1);
	}

	i = 0;
	while (i < extlstuse) {
	    j = extptrs[i];	 /*  Alphabetical order  */
	    fprintf(fout, "%s\n", extnlst[j].url);
	    i++;
	}
    
	fclose(fout);

	printf("reporturls(): Wrote %d urls to external urls report\n", i);
    }

    return(0);
}

/*  Process sitemap  */
int procsitemap(char *line, int *tags)
{
    int i, j, ret;
    char url[CGS_PATH_SIZ];

    i = 0;
    j = 0;
    ret = 0;
    memset(url, 0, CGS_PATH_SIZ);

    while (line[i] != 0 && i < CGS_BUF_SIZ) {
	if (strncasecmp(line + i, "<sitemapindex", 13) == 0) {
	    /*  Sitemap that lists Sitemaps  */
	    if(debflg != 0)
		fprintf(stderr, "Debug: procsitemap(): Sitemap-Index\n");
	    *tags = CGS_TAG_STMPIDX;	/*  Reset other tags  */
	    i += 12;
	} else if (strncasecmp(line + i, "<urlset", 7) == 0) {
	    /*  Sitemap that lists Urls  */
	    if(debflg != 0)
		fprintf(stderr, "Debug: procsitemap(): Url-Set\n");
	    *tags = CGS_TAG_URLSET;	/*  Reset other tags  */
	    i += 6;
	} else if (*tags == CGS_TAG_STMPIDX && \
	  strncasecmp(line + i, "<sitemap>", 9) == 0) {
	    /*  Sitemap follows  */
	    *tags |= CGS_TAG_SITEMAP;
	    i += 8;
	} else if (*tags == CGS_TAG_URLSET && \
	  strncasecmp(line + i, "<url>", 5) == 0) {
	    /*  Url follows  */
	    *tags |= CGS_TAG_URL;
	    i += 4;
	//} else if ((*tags & CGS_TAG_URL) != 0 && 
	} else if ((*tags & CGS_TAG_SIUR) != 0 && \
	  strncasecmp(line + i, "<loc>", 5) == 0) {
	    /*  Loc in Sitemap or Url  */
	    *tags |= CGS_TAG_LOC;
	    i += 4;
	} else if ((*tags & CGS_TAG_LOC) != 0 && \
	  strncasecmp(line + i, "</loc>", 6) == 0) {
	    /*  End of Loc  */
	    if (j != 0) {
		/*  Add loc to list  */
		if(debflg != 0)
		    fprintf(stderr, "Debug: procsitemap(): Loc: %s\n", url);
		ret = findurl(url, 0);
		if ((*tags & CGS_TAG_SITEMAP) != 0) {
		    /*  Loc is Sitemap; Check  */
		    if (j < 4 || strncasecmp(url + j - 4, ".xml", 4) != 0) {
			/*  May be gzipped  */
			urlist[ret].type = CGS_TYP_OTHER;
		    } else {
			/*  Is Xml; Fetch  */
			urlist[ret].type = CGS_TYP_SITEMAP;
		    }
		}
		j = 0;
		url[0] = 0;
	    }
	    *tags &= (~ CGS_TAG_LOC);
	    i += 5;
	} else if (strncasecmp(line + i, "</sitemap>", 10) == 0) {
	    /*  Clear Loc and Sitemap  */
	    *tags &= (~ CGS_TAG_LOSI);
	    i += 9;
	} else if (strncasecmp(line + i, "</url>", 6) == 0) {
	    /*  Clear Loc and Url  */
	    *tags &= (~ CGS_TAG_LOUR);
	    i += 5;
	} else if (strncasecmp(line + i, "</sitemapindex>", 15) == 0) {
	    *tags = 0;	/*  Reset tags  */
	    i += 14;
	} else if (strncasecmp(line + i, "</urlset>", 9) == 0) {
	    *tags = 0;	/*  Reset tags  */
	    i += 8;
	} else if ((*tags & CGS_TAG_LOC) != 0) {
	    /*  Add to Loc  */
	    url[j] = line[i];
	    j++;
	    url[j] = 0;
	    if (j > CGS_PATH_MAX2) {
		fprintf(stderr, "procsitemap(): Url too long\n");
		j = 0;
		url[0] = 0;
	    }
	}
	i++;
    }

    return(0);
}

/*  Process robots.txt  */
int procrobots(char *line)
{
    int len, ret;
    char url[CGS_PATH_SIZ];

    len = 0;
    ret = 0;
    memset(url, 0, CGS_PATH_SIZ);

    if (strncasecmp(line, "Sitemap: ", 9) == 0) {
	/*  Found Sitemap  */
	len = strlen(line);
	if (len < CGS_PATH_SIZ) {
	    sscanf(line, "%*s %s", url);
	    len = strlen(url);
	    if (len > 4 && strcmp(url + len - 4, ".xml") == 0) {
		if (debflg != 0)
		    fprintf(stderr, "Debug: procrobots(): Url %s\n", url);
		ret = findurl(url, 0);
		urlist[ret].type = CGS_TYP_SITEMAP;
	    }
	} else {
	    fprintf(stderr, "procrobots(): Line too long\n");
	}
    }

    return(0);
}

/*  Find a substring in a string  */
int findstr(char *s1, char *s2, int start, int len)
{
    int i;

    i = start;
    while (s1[i] != 0) {
	if (strncasecmp(s1 + i, s2, len) == 0)
	    return(i);
	i++;
	if (i > CGS_BUF_MAX2)
	    return(0);
    }

    return(0);
}

/*  Process HEAD  */
int prochead(struct urlprps *urlprops)
{
    char loc[CGS_PATH_SIZ];
    int len, respcod, ret, type;
    char *base, *line, *mhost, *url;
    int  docnt, lincnt, outp;

    len = 0;
    respcod = 0;
    ret = 0;
    memset(loc, 0, CGS_PATH_SIZ);

    /*  From urlprops  */
    base   = urlprops->base;
    docnt  = urlprops->docnt;
    line   = urlprops->line;
    lincnt = urlprops->linecnt;
    mhost  = urlprops->mhost;
    outp   = urlprops->outp;

    /*  From urlist  */
    type   = urlist[docnt].type;	/*  Store type if pre-set  */
    url    = urlist[docnt].url;

    if (line[0] == 10 || (line[0] == 13 && line[1] == 10)) {
	/*  End of header  */
	if(debflg != 0)
	    fprintf(stderr, "Debug: prochead(): End of header\n");
	/*  prochtpres() checks return if brokflg != 0  */
	return(1);
    } else if (lincnt == 0) {
	sscanf(line, "%*s %d %*s", &respcod);
	urlist[docnt].stat = (unsigned short) respcod;
    } else if (strncasecmp(line, "Location: ", 10) == 0 && \
	outp < CGS_PATH_SIZ + 10) {
	    /*  Redirect  */
	    strcpy(loc, line + 10);
	    len = outp - 10;
	    if(len < 0)
		len = 0;
	    if (len > 0 && loc[len] == 10) {
		loc[len] = 0;
		len--;
	    }
	    if (len > 0 && loc[len] == 13) {
		loc[len] = 0;
		len--;
	    }
	    if (len > 10) {
		/* procpath(mhost, base, link); */
		ret = procpath(mhost, base, loc);
		if (ret > -1) {
		    /*  Store redirect num  */
		    urlist[docnt].redir = \
		      (unsigned short) (ret + 1);
		    if (urlist[ret].type == 0) {
			/*  No type yet; Pre-set type if sitemap or robots  */
			if (type == CGS_TYP_ROBOTS)
			    urlist[ret].type = CGS_TYP_ROBOTS;
			else if (type == CGS_TYP_SITEMAP)
			    urlist[ret].type = CGS_TYP_SITEMAP;
		    }
		}
	    }
    } else if (strncasecmp(line, "Content-Type: ", 14) == 0) {
	/*  Content-Type: text/html; charset=UTF-8  */
	if (strncasecmp(line + 14, "application/pdf", 15) == 0) {
	    urlist[docnt].type = CGS_TYP_PDF;
	} else if (strncasecmp(line + 14, "application/xml", 15) == 0) {
	    if (type != CGS_TYP_SITEMAP) {
		/*  If not pre-set, ignore  */
		urlist[docnt].type = CGS_TYP_OTHER;
	    }
	} else if (strncasecmp(line + 14, "text/", 5) != 0) {
	   urlist[docnt].type = CGS_TYP_OTHER;
	} else if (strncasecmp(line + 19, "html", 4) == 0) {
	    urlist[docnt].type = CGS_TYP_HTML;
	} else if (type != CGS_TYP_ROBOTS) {
	    urlist[docnt].type = CGS_TYP_TXT;
	}
	ret = findstr(line, "charset=", 19, 8);
	if (ret != 0 && findstr(line, "UTF-8", ret + 8, 5) == 0) {
	    fprintf(stderr, "prochead(): %s is not UTF-8\n", url);
	    strncpy(urlist[docnt].charset, line + ret + 8, CGS_CHARSET_MAX);
	    len = outp - ret - 8;
	    if (urlist[docnt].charset[len] == 10) {
		urlist[docnt].charset[len] = 0;
		len--;
	    }
	    if (urlist[docnt].charset[len] == 13) {
		urlist[docnt].charset[len] = 0;
		len--;
	    }
	    fprintf(stderr, "prochead(): Charset: %s\n", urlist[docnt].charset);
	    /*  Charset is often wrong: Use superset if available.  */
	    if(strcasecmp(urlist[docnt].charset, "ISO-8859-1") == 0)
		strcpy(urlist[docnt].charset, "Windows-1252");
	    else if(strcasecmp(urlist[docnt].charset, "ISO-8859-9") == 0)
		strcpy(urlist[docnt].charset, "Windows-1254");
	}
    } else if (strncasecmp(line, "Content-Encoding: ", 18) == 0) {
	urlist[docnt].type = CGS_TYP_OTHER;
    }

    return(0);
}

/*  Process charset found from http-equiv  */
int procharset(char *charset)
{
    int i;

    i = 0;
    while (i < CGS_CHARSET_SIZ && charset[i] != 0) {
	if (charset[i] == '\"') {
	    break;   
	}
	i++;
    }
    if (i > CGS_CHARSET_MAX) {
	fprintf(stderr, "procharset(): Invalid charset\n");
	charset[0] = 0;
	return(-1);
    }

    if (charset[i] == '\"') {
	/*  Terminate  */
	charset[i] = 0;
    } else {
	/*  Charset should end in '"'  */
	fprintf(stderr, "procharset(): Invalid charset\n");
	charset[0] = 0;
	return(-1);
    }

    i = 0;
    while (i < CGS_CHARSET_SIZ && charset[i] != 0) {
	if ((charset[i] < '0' && charset[i] != '-') || \
	  (charset[i] > '9' && charset[i] < 'A') || \
	  (charset[i] > 'Z' && charset[i] < 'a' && charset[i] != '_') || \
	  charset[i] > 'z') {
	    fprintf(stderr, "procharset(): Invalid charset\n");
	    charset[0] = 0;
	    return(-1);
	}
	i++;
    }

    /*  Charset is often wrong: Use superset if available.  */
    if(strcasecmp(charset, "UTF-8") == 0)
	charset[0] = 0;
    else if(strcasecmp(charset, "ISO-8859-1") == 0)
	strcpy(charset, "Windows-1252");
    else if(strcasecmp(charset, "ISO-8859-9") == 0)
	strcpy(charset, "Windows-1254");

    return(0);
}

/*  Process GET  */
int procget(struct urlprps *urlprops)
{
    int ret;
    char *base, *charset, *line, *line2, *line3, *mhost, *url;
    int *chrcnt, crawl, docnt, len, lincnt, *tags, type;
    struct titlprps *title;

    ret = 0;

    /*  From urlprops  */
    base    = urlprops->base;		/*  Base URL  */
    chrcnt  = urlprops->chrcnt;
    crawl   = urlprops->crawl;		/*  Follow links if 1  */
    docnt   = urlprops->docnt;
    len     = urlprops->outp + 1;
    lincnt  = urlprops->linecnt;
    line    = urlprops->line;
    line2   = urlprops->line2;
    line3   = urlprops->line3;
    mhost   = urlprops->mhost;		/*  Method://Host/  */
    tags    = &(urlprops->tags);
    title   = &(urlprops->title);

    /*  From urlist  */
    charset = urlist[docnt].charset;
    type    = urlist[docnt].type;
    url     = urlist[docnt].url;

    /* Type is never PDF or Other  */
    if (type == CGS_TYP_HTML) {
	/*  HTML  */
	if (charset[0] == 0) {
	    /*  UTF-8  */
	    ret = remtag(line, line3, len, mhost, base, tags, title, docnt, \
	      crawl);
	} else {
	    line2[0] = 0;
	    /*  int toutf(char *inbuf, int inlen, char *outbuf, int outsiz, \
	      char *charset)  */
	    ret = toutf(line, len, line2, CGS_BUF_MAX, charset);
	    if (ret > 0) {
		ret = remtag(line2, line3, ret, mhost, base, tags, title, \
		  docnt, crawl);
	    } else {
		fprintf(stderr, "procget(): toutf() failed\n");
	    }
	}
	if (ret < 0) {
	    fprintf(stderr, "procget(): Indexing of %s terminated at line %d\n", \
	      url, lincnt + 1);
	    return(-1);
	} else if (ret > 0) {
	    ret = findwords(line3, ret, docnt, chrcnt);
	}
    } else if (type == CGS_TYP_ROBOTS) {
	/*  robots.txt  */
	ret = procrobots(line);
    } else if (type == CGS_TYP_SITEMAP) {
	/*  Sitemap  */
	ret = procsitemap(line, tags);
    } else if (type == CGS_TYP_TXT) {
	/*  Text  */
	if (charset[0] == 0) {
	    /*  UTF-8  */
	    ret = findwords(line, len, docnt, chrcnt);
	} else {
	    line2[0] = 0;
	    ret = toutf(line, len, line2, CGS_BUF_MAX, charset);
	    if (ret > 0)
		ret = findwords(line2, ret, docnt, chrcnt);
	}
    } else {
	fprintf(stderr, "procget(): %s is not Text, Html or Sitemap\n", \
	  url);
    }
    if (ret != 0) {
	fprintf(stderr, "procget(): %s is not a UTF-8 file\n", url);
	fprintf(stderr, "procget(): Indexing of %s terminated at line %d\n", \
	  url, lincnt + 1);
	return(-1);
    }

    if(title->str[0] != 0)
	strcpy(urlist[docnt].title, title->str);

    return(0);
}

/*  HEAD to GET switchover  */
int head2get(struct urlprps *urlprops)
{
    int docnt;

    docnt = urlprops->docnt;

    if (prepget(urlprops) != 0 && urlist[docnt].type > CGS_TYP_SITEMAP) {
	/*  Unwanted Content-Type; Abort GET  */
	fprintf(stderr, "head2get(): Aborting Get %s\n", urlist[docnt].url);
	urlprops->cont = 0;
	return(-1);
    }

    return(0);
}

/*  Process HTTP result  */
static size_t prochtpres(void *contents, size_t size, size_t nmemb, void *userp)
{
    int docnt, i, ret;
    size_t realsize;
    char *str;
    struct urlprps *urlprops;

    i = 0;
    ret = 0;

    realsize = size * nmemb;
    urlprops = (struct urlprps *) userp;
    str = (char *) contents;
    docnt = urlprops->docnt;

    if (urlprops->cont == 0) {
	/*  Don't do anything  */
	return(realsize);
    }

    if (urlprops->mode != 1 || urlist[docnt].type < CGS_TYP_PDF) {
	while (i < (int) realsize) {
	    urlprops->line[urlprops->outp] = str[i];
	    if (urlprops->line[urlprops->outp] == 10) {
		urlprops->line[urlprops->outp + 1] = 0;	/*  Terminate  */
		if (urlprops->mode != 1) {
		    /*  Mode is 0 or 2; HEAD  */
		    ret = prochead(urlprops);
		    if (brokflg != 0 && ret != 0) {
			/*  Switch from HEAD to GET  */
			if (head2get(urlprops) != 0) {
			    /*  Unwanted Content-Type  */
			    break;
			}
		    }
		} else {
		    /*  Mode 1; GET  */
		    procget(urlprops);
		}
		urlprops->linecnt++;
		urlprops->outp = 0;
		urlprops->line[0] = 0;
	    } else {
		urlprops->outp++;
		if (urlprops->outp > CGS_BUF_MAX2) {
		    fprintf(stderr, "prochtpres(): Line too long\n");
		    urlprops->outp = 0;
		    urlprops->line[0] = 0;
		    return(realsize);
	        }
	    }
	    i++;
	}
    } else {
	/*  PDF  */
	if (debflg != 0) {
	    fprintf(stderr, "Debug: prochtpres(): write %d bytes to Pdf\n", \
	      (int) realsize);
	}
	if(fwrite(str, 1, realsize, urlprops->pdfil) < 1)
	    fprintf(stderr, "prochtpres(): Can't write to Pdf file\n");
    }

    return(realsize);
}

/*  Check Curl result code  */
int chkcurlres(CURLcode res)
{
    if (res != CURLE_OK) {
	fprintf(stderr, "chkcurlres(): curl_easy_setopt() failed: %s\n", \
	  curl_easy_strerror(res));
	return(-1);
    }

    return(0);
}

/*  Abort if coninue is zero  */
int chkarbt(void *clientp, curl_off_t dltotal, curl_off_t dlnow, \
  curl_off_t ultotal, curl_off_t ulnow)
{
    int *cont;

    cont = (int *) clientp;

    if (debflg != 0) {
	fprintf(stderr, "Debug: chkarbt(): Down: %ld %ld Up: %ld %ld\n", \
	  dltotal, dlnow, ultotal, ulnow);
    }

    if (cont != NULL) {
	if(debflg != 0)
	    fprintf(stderr, "Debug: chkarbt(): Cont %d\n", *cont);
	if (*cont == 0) {
	    return(1);
	} else {
	    return(0);
	}
    } else {
	return(0);
    }
}

/*  Do a LibCurl HEAD or a GET  */
int prochtp(struct urlprps *urlprops)
{
    char *url;
    int docnt, mode;
    CURLcode res;

    /*
    *  Modes
    *
    *  0  HEAD
    *  1  GET
    *  2  Use GET to emulate HEAD
    *
    */

    docnt = urlprops->docnt;
    mode  = urlprops->mode;
    res   = 0;
    url   = urlist[docnt].url;

    if (debflg != 0) {
	if(mode == 0)
	    fprintf(stderr, "Debug: prochtp(): Head %s\n", url);
	else
	    fprintf(stderr, "Debug: prochtp(): Get %s\n", url);
    }

    //if (mode != 0) {
	/*  GET; Find base URL and Method://Host/  */
	//fndbasurl(url, urlprops->base, urlprops->mhost);
    //}
    /*  We need these in case of a redirect  */
    fndbasurl(url, urlprops->base, urlprops->mhost);

    /*  Reset  */
    curl_easy_reset(curlh);

    /*  URL  */
    res = curl_easy_setopt(curlh, CURLOPT_URL, url);
    if (chkcurlres(res) != 0)
	return(-1);

    if (brokflg != 0 && mode == 2) {
	/*  Enable abort  */
	res = curl_easy_setopt(curlh, CURLOPT_NOPROGRESS, 0L);
	if(chkcurlres(res) != 0)
	    return(-1);
	res = curl_easy_setopt(curlh, CURLOPT_XFERINFODATA, \
	  (void *) &(urlprops->cont));
	if(chkcurlres(res) != 0)
	    return(-1);
	res = curl_easy_setopt(curlh, CURLOPT_XFERINFOFUNCTION, chkarbt);
	if(chkcurlres(res) != 0)
	    return(-1);
    }

    /*  Buffer size  */
    res = curl_easy_setopt(curlh, CURLOPT_BUFFERSIZE, CGS_CURL_BUF_SIZ);
    if(chkcurlres(res) != 0)
	return(-1);
    /*  HTTP Version  */
    res = curl_easy_setopt(curlh, CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_0);
    if(chkcurlres(res) != 0)
	return(-1);
    /*  User agent  */
    res = curl_easy_setopt(curlh, CURLOPT_USERAGENT, CGS_USER_AGENT);
    if(chkcurlres(res) != 0)
	return(-1);
    /*  Connect timeout  */
    res = curl_easy_setopt(curlh, CURLOPT_CONNECTTIMEOUT, CGS_CURL_CON_TIM);
    if(chkcurlres(res) != 0)
	return(-1);
    /*  Overal timeout  */
    res = curl_easy_setopt(curlh, CURLOPT_TIMEOUT, CGS_CURL_TIMEOUT);
    if(chkcurlres(res) != 0)
	return(-1);

    if (strncmp(url, "https://", 8) == 0) {
	/*  No cert check  */
	res = curl_easy_setopt(curlh, CURLOPT_SSL_VERIFYPEER, 0L);
	if(chkcurlres(res) != 0)
	    return(-1);
    }

    if (mode != 1) {
	/*  HEAD  */
	res = curl_easy_setopt(curlh, CURLOPT_HEADER, 1L);
	if(chkcurlres(res) != 0)
	    return(-1);
	if (brokflg == 0 || mode == 0) {
	    /*  No body  */
	    res = curl_easy_setopt(curlh, CURLOPT_NOBODY, 1L);
	    if(chkcurlres(res) != 0)
		return(-1);
	}
	if (strncmp(url, "https://", 8) == 0) {
	    /*  No proxy for https HEAD  */
	    res = curl_easy_setopt(curlh, CURLOPT_NOPROXY, "*");
	    if(chkcurlres(res) != 0)
		return(-1);
	}
    }

    /*  How to process the data  */
    res = curl_easy_setopt(curlh, CURLOPT_WRITEFUNCTION, prochtpres);
    if(chkcurlres(res) != 0)
	return(-1);
    res = curl_easy_setopt(curlh, CURLOPT_WRITEDATA, (void *) urlprops);
    if(chkcurlres(res) != 0)
	return(-1);

    /* Perform the request */
    res = curl_easy_perform(curlh);
    if (res != CURLE_OK) {
	fprintf(stderr, "prochtp(): curl_easy_perform() failed: %s\n", \
	  curl_easy_strerror(res));
        return(-1);
    }

    return(0);
}

/*  Index PDF file  */
int procpdf(char *pdfnam, int *chrcnt, int docnt)
{
    char comnd[CGS_COM_SIZ], line[CGS_BUF_SIZ];
    int len;
    FILE *p1;

    memset(comnd, 0, CGS_COM_SIZ);
    memset(line,  0, CGS_BUF_SIZ);
    p1 = NULL;

    snprintf(comnd, CGS_COM_MAX, "%s %s -", CGS_PDF_TO_TEXT, pdfnam);
    if(debflg != 0)
	fprintf(stderr, "Debug: procpdf(): %s\n", comnd);
    if ((p1 = popen(comnd, "r")) == NULL) {
	fprintf(stderr, "procpdf(): Can't open pdftotext\n");
	return(-1);
    }

    while (fgets(line, CGS_BUF_SIZ, p1)) {
	len = strlen(line);
	if (debflg != 0) {
	    fprintf(stderr, "Debug: procpdf(): Read %d bytes from pdftotext\n", \
	      len);
	    //fprintf(stderr, "%s", line);
	}
	if (findwords(line, len, docnt, chrcnt) != 0) {
	    fprintf(stderr, "procpdf(): %s is not a UTF-8 file, ", pdfnam);
	    fprintf(stderr, "Indexing terminated\n");
	    pclose(p1);
	    return(-1);
	}
    }

    pclose(p1);

    return(0);
}

/*  Prepare PDF proccessing  */
FILE * prepdf(int docnt, char *pdfnam)
{
    FILE *pdfil;

    pdfil = NULL;

    #ifdef CGS_DO_TEST
     sprintf(pdfnam, "%04X.pdf", urlist[docnt].num);
    #else
     snprintf(pdfnam, CGS_PDF_MAX, "%snew/%04X.pdf", CGS_PATH, \
       urlist[docnt].num);
    #endif
    pdfil = fopen(pdfnam, "w");

    if (pdfil != NULL) {
	if (debflg != 0) {
	    fprintf(stderr, "Debug: prepdf(): Get: %s Type: Pdf\n", \
	      urlist[docnt].url);
	    fprintf(stderr, "Debug: prepdf(): save to: %s\n", pdfnam);
	}
    } else {
	fprintf(stderr, "prepdf(): can't open %s\n", pdfnam);
	return(NULL);
    }

    return(pdfil);
}

/*  Prepare for get  */
int prepget(struct urlprps *urlprops)
{
    char *pdfnam;
    int  docnt;

    docnt  = urlprops->docnt;
    pdfnam = urlprops->pdfnam;

    if (urlist[docnt].redir == 0 && urlist[docnt].stat == 200 && \
      urlist[docnt].type < CGS_TYP_OTHER) {
	urlprops->linecnt = 0;
	if (pdflg != 0 && urlist[docnt].type == CGS_TYP_PDF) {
	    /*  Is a PDF file  */
	    if (debflg != 0) {
		fprintf(stderr, "Debug: prepget(): Get: %s Type: %s\n", \
		  urlist[docnt].url, typtxts[urlist[docnt].type]);
	    }
	    urlprops->pdfil = prepdf(docnt, pdfnam);
	    if (urlprops->pdfil != NULL ) {
		/*  PDF file opened  */
		if (urlprops->mode == 1) {
		    /*  Set to 1 by procurlst()  */
		    prochtp(urlprops);
		} else {
		    /*  Was 2; prochtp() already running  */
		    urlprops->mode = 1;
		}
	    } else {
		/*  Don't do anything  */
		return(-1);
	    }
	} else if (urlist[docnt].type < CGS_TYP_PDF) {
	    /*  Is Text, HTML or Sitemap  */
	    if (debflg != 0) {
		fprintf(stderr, "Debug: prepget(): Get: %s Type: %s\n", \
		  urlist[docnt].url, typtxts[urlist[docnt].type]);
	    }
	    if (urlprops->mode == 1) {
		prochtp(urlprops);
	    } else {
		/*  Was 2; prochtp() already running  */
		urlprops->mode = 1;
	    }
	} else {
	    /*  Don't do anything  */
	    return(-1);
	}
    } else {
	/*  Don't do anything  */
	return(-1);
    }

    return(0);
}

/*  Process URL list  */
int procurlst(int *chrcnt, int crawl, int *docnt, struct timespec pause)
{
    char *pdfnam;
    struct timespec rem;
    struct urlprps urlprops;		/*  Anything about current URL  */

    memset(&rem, 0, sizeof(rem));
    memset(&urlprops, 0, sizeof(urlprops));

    pdfnam = urlprops.pdfnam;

    /*  Crawl these urls  */
    do {
	/*  Init URL properties  */
	memset(&urlprops, 0, sizeof(urlprops));
	urlprops.chrcnt = chrcnt;
	urlprops.cont   = 1;		/*  Continue  */
	urlprops.crawl  = crawl;	/*  Follow found links if 1  */
	urlprops.docnt  = *docnt;

	if (debflg != 0) {
	    fprintf(stderr, "Debug: procurlst(): Url: %04X %s\n", \
	      *docnt + 1, urlist[*docnt].url);
	}

	if (brokflg == 0) {
	    /*  HEAD, mode is zero  */
	    prochtp(&urlprops);
	    /*  GET  */
	    urlprops.mode = 1;
	    prepget(&urlprops);
	} else {
	    /*  Combined HEAD and GET  */
	    urlprops.mode = 2;
	    prochtp(&urlprops);
	}
	if (urlprops.pdfil != NULL) {
	    /*  There is a PDF file  */
	    fclose(urlprops.pdfil);
	    /*  Process PDF file  */
	    procpdf(pdfnam, chrcnt, *docnt);
	    if (debflg == 0 && remove(pdfnam) < 0) {
		/*  Don't remove file with debug on  */
		fprintf(stderr, "procurlst(): can't delete %s\n", pdfnam);
	    }
	    pdfnam[0] = 0;
	    urlprops.pdfil = NULL;
	}

	(*docnt)++;
	/*  Check maximum  */
	if (*docnt > 65530) {
	    fprintf(stderr, "procurlst(): Too many documents.\n");
	    freethelot(1);
	}
	/*  Check sizes  */
	if (absflg != 0 && *docnt > abstrmax) {
	    growabstr();
	}
	if (*docnt > urlstmax) {
	    if (debflg != 0) {
		fprintf(stderr, "procurlst(): urlstmax reached\n");
	    }
	    growurlst();
	}
	/*  Wait, if any  */
	if (pause.tv_sec != 0 || pause.tv_nsec != 0) {
	    if (nanosleep(&pause, &rem) != 0) {
		fprintf(stderr, "procurlst(): nanosleep() error\n");
		if(errno == EINTR)
		    nanosleep(&rem, &rem);
	    }
	}

	/*  urlstuse points to next record  */
    } while (*docnt < urlstuse);

    return(*docnt);
}

/*  Get sitemap  */
int getsitemap(int urlnr, struct timespec pause)
{
    struct urlprps urlprops;
    struct timespec rem;

    memset(&rem, 0, sizeof(rem));
    memset(&urlprops, 0, sizeof(urlprops));

    urlprops.cont = 1;	/*  Continue  */
    urlprops.docnt = urlnr;

    urlist[urlnr].type = CGS_TYP_SITEMAP;

    if (debflg != 0) {
	fprintf(stderr, "Debug: getsitemap(): Url: %04X %s\n", \
	  urlnr + 1, urlist[urlnr].url);
    }

    if (brokflg == 0) {
	/*  HEAD, mode is zero  */
	prochtp(&urlprops);
	/*  GET  */
	urlprops.mode = 1;
	prepget(&urlprops);
    } else {
	/*  Combined HEAD and GET  */
	urlprops.mode = 2;
	prochtp(&urlprops);
    }

    /*  Wait, if any  */
    if (pause.tv_sec != 0 || pause.tv_nsec != 0) {
	if (nanosleep(&pause, &rem) != 0) {
	    fprintf(stderr, "procurlst(): nanosleep() error\n");
            if(errno == EINTR)
		nanosleep(&rem, &rem);
	}
    }

    return(urlstuse);
}

/* Get robots.txt  */
int getrobots(int urlnr, struct timespec pause)
{
    struct urlprps urlprops;
    struct timespec rem;

    memset(&rem, 0, sizeof(rem));
    memset(&urlprops, 0, sizeof(urlprops));

    urlprops.cont = 1;  /*  Continue  */
    urlprops.docnt = urlnr;

    urlist[urlnr].type = CGS_TYP_ROBOTS;

    if (debflg != 0) {
	fprintf(stderr, "Debug: getrobots(): Url: %04X %s\n", \
	  urlnr + 1, urlist[urlnr].url);
    }

    if (brokflg == 0) {
	/*  HEAD, mode is zero  */
	prochtp(&urlprops);
	/*  GET  */
	urlprops.mode = 1;
	prepget(&urlprops);
    } else {
	/*  Combined HEAD and GET  */
	urlprops.mode = 2;
	prochtp(&urlprops);
    }

    /*  Wait, if any  */
    if (pause.tv_sec != 0 || pause.tv_nsec != 0) {
	if (nanosleep(&pause, &rem) != 0) {
	    fprintf(stderr, "procurlst(): nanosleep() error\n");
	    if(errno == EINTR)
		nanosleep(&rem, &rem);
	}
    }

    return(urlstuse);
}

/*  Compare start URLs  */
int compstart(const void *r1, const void *r2)
{
    struct refurl *ref1, *ref2;

    ref1 = (struct refurl *) r1;
    ref2 = (struct refurl *) r2;

    return(strcmp(ref1->url, ref2->url));
}

/*  Process Ref URL list  */
int procreflst(char *listn, struct timespec pause)
{
    int chrcnt, cururl, docnt, i, len, lines, ret, size;
    char line[CGS_BUF_SIZ], url[CGS_PATH_SIZ];
    FILE *listf;

    chrcnt = 0;
    cururl = 0;
    docnt = 0;
    i = 0;
    len = 0;
    lines = 0;
    refurluse = 0;				/*  Nr of refurls  */
    ret = 0;
    size = 0;
    listf = NULL;
    //refurltab = NULL;	Moved to main()
    memset(line, 0, CGS_BUF_SIZ);
    memset(url, 0, CGS_PATH_SIZ);

    /*
    *  urlstentr   Space available for entries
    *  urlstmax    Two smaller!
    *  urlstsiz    Size in bytes
    *  urlstuse    Nr of urls in urlist
    *
    *  docnt       URL beeing proccessed
    */

    if ((listf = fopen(listn, "r")) == NULL) {
	fprintf(stderr, "procreflst(): Can't open %s\n", listn);
	freethelot(1);
    }

    while (fgets(line, CGS_BUF_SIZ, listf)) {
	if (line[0] == 10 || line[0] == 13 || line[0] == '#')
	    continue;
	lines++;
    }
    if (lines >= CGS_DFLT_DOCS - 2) {
	fprintf(stderr, "procrefst(): Too many ref urls\n");
	freethelot(1);
    }

    size = (lines + 1) * sizeof(struct refurl);
    refurltab = malloc(size);
    if (refurltab == NULL) {
	fprintf(stderr, "procreflst(): Can't alloc memory for refurltab\n");
	freethelot(1);
    }
    memset(refurltab, 0, size);
    if (debflg != 0) {
	fprintf(stderr, \
	  "Debug: procreflst(): allocated %d bytes for refurltab\n", \
	  size);
    }
    rewind(listf);
    while (fgets(line, CGS_BUF_SIZ, listf)) {
	if (line[0] == 10 || line[0] == 13 || line[0] == '#')
	    continue;
	len = strlen(line) - 1;
	if (line[len] == 10) {
	    line[len] = 0;
	    len--;
	}
	if (len == 0)
	    continue;
	if (line[len] == 13) {
	    line[len] = 0;
	    len--;
	}
	if (len == 0)
	    continue;

	if (strncmp(line, "http", 4) != 0) {
	    fprintf(stderr, "procreflst(): Not a Http Url: %s\n", line);
	    continue;
	}
	if (len > CGS_MHOST_MAX2) {
	    fprintf(stderr, "procreflst(): Refurl too long\n");
	    continue;
	}
	if (line[len] != '/') {
	    fprintf(stderr, "procreflst(): Apending slash to: %s\n", line);
	    len++;
	    line[len] = '/';
	    line[len + 1] = 0;
	}
	len++;

	strcpy(refurltab[refurluse].url, line);
	refurltab[refurluse].len = len;

	refurluse++;
	if (refurluse >= lines || refurluse >= CGS_DFLT_DOCS - 2)
	    break;
    }
    fclose(listf);

    if (debflg != 0) {
	i = 0;
	fprintf(stderr, \
	  "Debug: procreflst(): Ref Url list %d entries:\n", refurluse);
	while (i < refurluse) {
	    fprintf(stderr, "%04X %d %s\n", \
	      i + 1, refurltab[i].len, refurltab[i].url);
	    i++;
	}
    }

    /*  Sort reference / start URL list  */
    qsort(refurltab, refurluse, sizeof(struct refurl), compstart);

    printf("procreflst(): Will crawl these sites:\n");
    for(i = 0; i < refurluse; i++)
	printf("%s\n", refurltab[i].url);
    printf("\n");

    /*
    *  Copy URL of first site to crawl to urlist
    *  Crawl this site
    *  Copy URL of second site to crawl to urlist
    *  Crawl this site
    *  Etc ...
    *
    *  findurl() makes sure each URL is in the list only once.
    */

    i = 0;
    urlstuse = 0;
    while (i < refurluse) {
	if (mapflg != 0) {
	    /*  Try sitemap first  */
	    cururl = urlstuse;	/*  Current Url  */
	    sprintf(url, "%ssitemap.xml", refurltab[i].url);
	    //sprintf(url, "%stest-sitemap.xml", refurltab[i].url);
	    ret = findurl(url, 0);
	    if (ret >= cururl) {
		/*  New Url */
		getsitemap(ret, pause);
		docnt++;
		if (urlstuse > docnt) {
		    if (debflg != 0) {
			fprintf(stderr, \
			  "Debug: procreflst(): Found sitemap\n");
		    }
		    /*  Index these links  */
		    docnt = procurlst(&chrcnt, 0, &docnt, pause);
		    /*  Next refurl  */
		    i++;
		    continue;
		} else {
		    /*  Try robots.txt  */
		    cururl = urlstuse;	/*  Current Url  */
		    sprintf(url, "%srobots.txt", refurltab[i].url);
		    //sprintf(url, "%stest-robots.txt", refurltab[i].url);
		    ret = findurl(url, 0);
		    if (ret >= cururl) {
			/*  New Url  */
			getrobots(ret, pause);
			docnt++;
			if (urlstuse > docnt) {
			    if (debflg != 0) {
				fprintf(stderr, \
				  "Debug: procreflst(): Found sitemap in robots.txt\n");
			    }
			    /*  Index these links  */
			    docnt = procurlst(&chrcnt, 0, &docnt, pause);
			    /*  Next refurl  */
			    i++;
			    continue;
			}
		    }
		}	/*  End of try robots  */
	    }
	}		/*  End of try sitemap  */

	/*  Copy start URL to list by means of findurl() */
	if (findurl(refurltab[i].url, 0) < docnt) {
	    // XXX if (debflg != 0) {
		fprintf(stderr, "procreflst(): %s already in list\n", \
		  refurltab[i].url);
	    //}
	    i++;
	    continue;
	}

	/*  Crawl this site  */
	docnt = procurlst(&chrcnt, 1, &docnt, pause);

	i++;
    }

    printf("procreflst(): Found %d words in %d documents\n", \
      wordlstuse, docnt);
    printf("procreflst(): Max wordlength is %d bytes\n", chrcnt);

    return(docnt);
}

#endif
/*  END of HTTP specific bits  */


int main(int argc, char **argv)
{
    int docnt, i, opt, ret, statflg, txtflg, wait;
    char listn[CGS_PATH_SIZ];
    struct timespec pause;

    absflg = 0;			/*  Generate abstracts file when 1  */
    abstrentr = 0;
    abstrmax = 0;
    abstrsiz = 0;
    #ifdef CGS_WITH_HTTP
     brokflg = 0;		/*  Broken server  */
     crawlflg = 0;		/*  Crawl when 1  */
    #endif
    compflg = 0;		/*  Compact wordlist file format when 1  */
    debflg = 0;			/*  Print debug if 1  */
    docnt = 0;			/*  Nr of documents  */
    #ifdef CGS_WITH_HTTP
     extnflg = 0;		/*  Export external links when 1  */
    #endif
    i = 0;
    lrgflg = 0;			/*  Use 32 bit word nrs if 1  */
    #ifdef CGS_WITH_HTTP
     mapflg = 0;		/*  Use sitemap if 1  */
    #endif
    opt = 0;
    #ifdef CGS_WITH_HTTP
     pdflg = 0;			/*  Process PDF when 1  */
    #endif
    reflg = 0;			/*  Use old wordlist if 1  */
    ret = 0;
    statflg = 0;		/*  Print word stats if 1  */
    synflg = 0;			/*  Use synonyms when non zero  */
    synentr = 0;		/*  Nr of entries in synonym table  */
    synlstsiz = 0;		/*  Size of synonym list  */
    txtflg = 0;			/*  Output Text if 1  */
    utflg = 0;			/*  Process non-ASCII if 1  */
    wordlstentr = 0;		/*  Space available for entries  */
    wordlstmax = 0;		/*  One smaller  */
    wordlstsiz = 0;		/*  Size in bytes  */
    wordlstuse = 0;		/*  Nr of words in wordlist  */
    wrdidxsiz = 0;		/*  Compact word list index size */
    memset(listn, 0, CGS_PATH_SIZ);	/*  List file name  */
    memset(&pause, 0, sizeof(pause));
    abstrlist = NULL;		/*  Per doc abstract list  */
    compwordlst = NULL;		/*  Compact word list  */
    docidx = NULL;		/*  Word to doc index  */
    synlstadr = NULL;		/*  Memory address of synonym list  */
    syntab = NULL;		/*  Synonym table  */
    wordlist = NULL;		/*  Word list / Word array  */
    wrdlstidx = NULL;		/*  Compact word list index  */

    wordptrsiz = 0;
    wrdptrs = NULL;

    #ifdef CGS_WITH_HTTP
     wait = 1000;		/*  Wait time between urls (ms)  */
     refurluse  = 0;		/*  Nr of URLs in start list  */
     urlstentr  = 0;		/*  Space available for entries  */
     urlstmax   = 0;		/*  Two smaller!  */
     urlstsiz   = 0;		/*  Size in bytes  */
     urlstuse   = 0;		/*  Nr of urls in urlist  */
     urlist  = NULL;		/*  URL list / URL array  */
     extlstentr = 0;
     extlstmax  = 0;
     extlstsiz  = 0;
     extlstuse  = 0;
     extnlst = NULL;
     curlh     = NULL;		/*  Curl handle  */
     refurltab = NULL;		/*  Start list  */
    #else
     wait = 0;			/*  Wait time between files  */
    #endif

    #ifdef CGS_WITH_HTTP
     while ((opt = getopt(argc, argv, "abcdef:hklmoprstuvw:")) != -1) {
    #else
     while ((opt = getopt(argc, argv, "adf:hklorstuvw:")) != -1) {
    #endif
	switch (opt) {
	    case 'a':
		absflg = 1;
		break;
	    #ifdef CGS_WITH_HTTP
	     case 'b':
		 brokflg = 1;
		 break;
	     case 'c':
		 crawlflg = 1;
		 break;
	    #endif
	    case 'd':
		debflg = 1;
		break;
	    #ifdef CGS_WITH_HTTP
	     case 'e':
		 extnflg = 1;
		 break;
	    #endif
	    case 'f':
		strncpy(listn, optarg, CGS_PATH_MAX);
		break;
	    case 'h':
		hlp();
		exit(0);
	    case 'k':
		compflg = 1;
		break;
	    case 'l':
		lrgflg = 1;
		break;
	    #ifdef CGS_WITH_HTTP
	     case 'm':
		 mapflg = 1;
		 break;
	    #endif
	    case 'o':
		synflg = 1;
		break;
	    #ifdef CGS_WITH_HTTP
	     case 'p':
		 pdflg = 1;
		 break;
	    #endif
	    case 'r':
		reflg = 1;
		break;
	    case 's':
		statflg = 1;
		break;
	    case 't':
		txtflg = 1;
		break;
	    case 'u':
		utflg  = 1;
		break;
	    case 'v':
		prtvers();
		exit(0);
	    case 'w':
		/*  Avoid rounding errors  */
		wait = ((int) (10000.0 * atof(optarg))) / 10;
		if (wait > 10000 || wait < 0) {
		    fprintf(stderr, "Wait should be 0 ... 10\n");
		    exit(1);
		}
		break;
	    default:
		hlp();
		exit(1);
	}
    }

    #ifdef CGS_WITH_HTTP
     /*  Check broken server  */
     if (brokflg != 0 && crawlflg == 0) {
	 fprintf(stderr, \
	   "main(): Broken server only works with crawl.\n");
	 exit(1);
     }
     /*  Check Use sitemap  */
     if (mapflg != 0 && crawlflg == 0) {
	 fprintf(stderr, \
	   "main(): Use sitemap only works with crawl.\n");
	 exit(1);
     }
     /*  Check PDF  */
     if (pdflg != 0 && crawlflg == 0) {
	 fprintf(stderr, \
	   "main(): Pdf indexing only works with crawl.\n");
	 exit(1);
     }
    #endif

    /*  Check file list file  */
    if (listn[0] == 0) {
	fprintf(stderr, \
	  "main(): Must specify list of files to be indexed '-f'.\n");
	exit(1);
    }

    /*  Convert wait to pause  */
    if (wait != 0) {
	pause.tv_sec = (time_t) (wait / 1000);
	pause.tv_nsec = (long) ((wait % 1000) * 1000000) ;
    }

    /*  Set to UTF-8  */
    setlocale(LC_ALL, "C.UTF-8");

    /*  Initialize  */
    init();

    if (synflg != 0) {
	/*  Initialize synonyms  */
	if (initsyn() != 0) {
	    fprintf(stderr, "main(): initsyn failed: Not using synonyms\n");
	    synflg = 0;
	}
    }

    #ifdef CGS_WITH_HTTP
     if(crawlflg == 0)
	 docnt = procfilst(listn, pause);
     else
	 docnt = procreflst(listn, pause);
    #else
     docnt = procfilst(listn, pause);
    #endif

    /*
    *  Dump the data
    *  Use ret as return value of main()
    */
    ret = 0;
    if (txtflg != 0) {
	/*  Text  */
	if(dumpwordlist(1) != 0)
	    ret = 1;
    } else {
	if(dumpwordlist(0) != 0)
	    ret = 1;
    }
    if(dumpindex(txtflg) != 0)
	ret = 1;
    if (absflg != 0) {
	if(dumpabstr(docnt, txtflg) != 0)
	    ret = 1;
    }
    #ifdef CGS_WITH_HTTP
     if (crawlflg != 0) {
	 if(dumpurlst(docnt, txtflg) != 0)
	     ret = 1;
	 if(reporturls() != 0)
	     ret = 1;
     }
    #endif
    #ifdef CGS_DO_TEST
     if (debflg != 0 && txtflg != 0) {
	 /*  Dump word pointers  */
	 dumpwrdptrs(lrgflg);
     }
    #endif
	
    if(statflg != 0)
 	countwords();	/*  Word frequency  */

    /*  Free the later allocated memory  */
    i = 0;
    while (i < wordlstentr) {
	if (docidx[i].siz > CGS_DFLT_DAT_SIZ) {
	    /*  Allocated later  */
	    free(docidx[i].docs);
	}
	i++;
    }

    /*  Free  */
    #ifdef CGS_WITH_HTTP
     curl_easy_reset(curlh);
     curl_easy_cleanup(curlh);
     curl_global_cleanup();
     if(extnlst != NULL)
	 free(extnlst);
     if(refurltab != NULL)
	 free(refurltab);
     free(urlist);
    #endif
    if(abstrlist != NULL)
	free(abstrlist);
    free(docidx);
    free(wrdptrs);
    free(wordlist);
    if(wrdlstidx != NULL)
	free(wrdlstidx);
    if(compwordlst != NULL)
	free(compwordlst);

    /*  Synonym list  */
    if(syntab != NULL)
	free(syntab);
    if(synlstadr != NULL)
	munmap(synlstadr, synlstsiz);
    if(synlstfd != 0)
	close(synlstfd);

    fflush(stderr);
    fflush(stdout);

    return(ret);
}

